/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2022. All rights reserved.
 */

package gmssl

/*
#include <openssl/evp.h>
*/
import "C"

import (
	"fmt"
	"runtime"
	"unsafe"
)

// GetDigestLength get digest length
func GetDigestLength(name string) (int, error) {
	cname := C.CString(name)
	defer C.free(unsafe.Pointer(cname))
	md := C.EVP_get_digestbyname(cname)
	if md == nil {
		return 0, GetErrors()
	}
	return int(C.EVP_MD_size(md)), nil
}

// GetDigestBlockSize get digest block size
func GetDigestBlockSize(name string) (int, error) {
	cname := C.CString(name)
	defer C.free(unsafe.Pointer(cname))
	md := C.EVP_get_digestbyname(cname)
	if md == nil {
		return 0, GetErrors()
	}
	return int(C.EVP_MD_block_size(md)), nil
}

// DigestContext digest context
type DigestContext struct {
	ctx *C.EVP_MD_CTX
}

// NewDigestContext new digest context
func NewDigestContext(name string) (*DigestContext, error) {
	cname := C.CString(name)
	defer C.free(unsafe.Pointer(cname))
	md := C.EVP_get_digestbyname(cname)
	if md == nil {
		return nil, GetErrors()
	}
	ctx := C.EVP_MD_CTX_new()
	if ctx == nil {
		return nil, GetErrors()
	}
	ret := &DigestContext{ctx: ctx}
	runtime.SetFinalizer(ret, func(ret *DigestContext) {
		C.EVP_MD_CTX_free(ret.ctx)
	})
	if C.EVP_DigestInit_ex(ctx, md, nil) != 1 {
		return nil, GetErrors()
	}
	return ret, nil
}

// Update update origin data
func (ctx *DigestContext) Update(data []byte) error {
	if len(data) == 0 {
		return nil
	}
	if C.EVP_DigestUpdate(ctx.ctx, unsafe.Pointer(&data[0]),
		C.size_t(len(data))) != 1 {
		return GetErrors()
	}
	return nil
}

// Final get final digest info
func (ctx *DigestContext) Final() ([]byte, error) {
	const evpErr = 0
	const tryTimes = 2
	maxHashLen := C.EVP_MAX_MD_SIZE * 2
	outbuf := make([]byte, maxHashLen)
	bufferAddr := unsafe.Pointer(&outbuf[0])
	outlen := C.uint(maxHashLen)

	for i := 1; i <= tryTimes; i++ {
		if C.EVP_DigestFinal_ex(ctx.ctx,
			(*C.uchar)(bufferAddr), &outlen) == evpErr {
			return nil, GetErrors()
		}
		if outlen <= C.EVP_MAX_MD_SIZE {
			return outbuf[:outlen], nil
		}
	}
	return nil, fmt.Errorf("try %d time, the outlen is not correct %d", tryTimes, maxHashLen)
}

// Reset reset digest ctx
func (ctx *DigestContext) Reset() error {
	md := C.EVP_MD_CTX_md(ctx.ctx)
	if md == nil {
		return GetErrors() // FIXME: return some errors
	}
	if C.EVP_DigestInit_ex(ctx.ctx, md, nil) != 1 {
		return GetErrors()
	}
	return nil
}
