/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2022. All rights reserved.
 */

package gmssl

/*
#include <stdlib.h>
#include <openssl/crypto.h>
#include <openssl/bio.h>
#include "gmcert.h"
#include "gmkey.h"
*/
import "C"

import (
	"errors"
	"runtime"
	"unsafe"
)

const sm2DigestLen = 32

// PublicKey encapsulate public key as type EVP_PKEY
type PublicKey struct {
	pkey *C.EVP_PKEY
}

// PrivateKey encapsulate private key as type EVP_PKEY
type PrivateKey struct {
	pkey *C.EVP_PKEY
}

// GeneratePrivateKey Generate private key by algorith and args
func GeneratePrivateKey(alg string, args [][2]string) (*PrivateKey, error) {
	calg := C.CString(alg)
	defer C.free(unsafe.Pointer(calg))

	ctx := C.new_pkey_keygen_ctx(calg, nil)
	defer C.EVP_PKEY_CTX_free(ctx)

	if ctx == nil {
		return nil, GetErrors()
	}
	var pkey *C.EVP_PKEY

	if alg == "DH" || alg == "DSA" {

		if 1 != C.EVP_PKEY_paramgen_init(ctx) {
			return nil, GetErrors()
		}
		err := procKeyGenerationArgs(args, ctx)
		if err != nil {
			return nil, err
		}
		if 1 != C.EVP_PKEY_paramgen(ctx, &pkey) {
			return nil, GetErrors()
		}
		if 1 != C.EVP_PKEY_keygen_init(ctx) {
			return nil, GetErrors()
		}
		if 1 != C.EVP_PKEY_keygen(ctx, &pkey) {
			return nil, GetErrors()
		}

	} else {
		if 1 != C.EVP_PKEY_keygen_init(ctx) {
			return nil, GetErrors()
		}

		err := procKeyGenerationArgs(args, ctx)
		if err != nil {
			return nil, err
		}

		if 1 != C.EVP_PKEY_keygen(ctx, &pkey) {
			return nil, GetErrors()
		}
	}

	sk := &PrivateKey{pkey: pkey}
	runtime.SetFinalizer(sk, func(sk *PrivateKey) {
		C.EVP_PKEY_free(sk.pkey)
	})

	return sk, nil
}

func procKeyGenerationArgs(args [][2]string, ctx *C.EVP_PKEY_CTX) error {
	for _, arg := range args {
		name := arg[0]
		value := arg[1]
		cName := C.CString(name)
		defer C.free(unsafe.Pointer(cName))
		cValue := C.CString(value)
		defer C.free(unsafe.Pointer(cValue))
		if C.EVP_PKEY_CTX_ctrl_str(ctx, cName, cValue) <= 0 {
			return GetErrors()
		}
	}
	return nil
}

// NewPrivateKeyFromPEM new private key from pem
func NewPrivateKeyFromPEM(pem string, pass string) (*PrivateKey, error) {
	cpem := C.CString(pem)
	defer C.free(unsafe.Pointer(cpem))
	cpass := C.CString(pass)
	defer C.free(unsafe.Pointer(cpass))
	bio := C.BIO_new_mem_buf(unsafe.Pointer(cpem), -1)
	if bio == nil {
		return nil, GetErrors()
	}
	defer C.BIO_free(bio)
	pkey := C.PEM_read_bio_PrivateKey(bio, nil, nil, unsafe.Pointer(cpass))
	if pkey == nil {
		return nil, GetErrors()
	}
	sk := &PrivateKey{pkey: pkey}
	runtime.SetFinalizer(sk, func(sk *PrivateKey) {
		C.EVP_PKEY_free(sk.pkey)
	})
	return sk, nil
}

// GetPublicKeyPEM get public key pem format value
func (sk *PrivateKey) GetPublicKeyPEM() (string, error) {
	bio := C.BIO_new(C.BIO_s_mem())
	if bio == nil {
		return "", GetErrors()
	}
	defer C.BIO_free(bio)
	if C.pem_write_bio_pubkey(bio, sk.pkey) != 1 {
		return "", GetErrors()
	}
	var p *C.char
	pubKeyLen := C.BIO_get_mem_data_func(bio, &p)
	if pubKeyLen <= 0 {
		return "", GetErrors()
	}
	return C.GoString(p)[:pubKeyLen], nil
}

// NewPublicKeyFromPEM new public key struct from pem format
func NewPublicKeyFromPEM(pem string) (*PublicKey, error) {
	cpem := C.CString(pem)
	defer C.free(unsafe.Pointer(cpem))
	bio := C.BIO_new_mem_buf(unsafe.Pointer(cpem), -1)
	if bio == nil {
		return nil, GetErrors()
	}
	defer C.BIO_free(bio)
	pkey := C.pem_read_bio_pubkey(bio)
	if pkey == nil {
		return nil, GetErrors()
	}
	pk := &PublicKey{pkey: pkey}
	runtime.SetFinalizer(pk, func(pk *PublicKey) {
		C.EVP_PKEY_free(pk.pkey)
	})
	return pk, nil
}

// GetDer get der format public key
func (pk *PublicKey) GetDer() ([]byte, error) {
	var der *C.uchar
	var length C.int
	defer func() {
		if der != nil {
			C.free(unsafe.Pointer(der))
		}
	}()
	length = C.i2d_PublicKey(pk.pkey, &der)
	if length < 1 {
		return nil, errors.New("get der failed for public key")
	}
	goDer := C.GoBytes(unsafe.Pointer(der), length)
	return goDer, nil
}

// GetPEM get pem format value of public key
func (pk *PublicKey) GetPEM() (string, error) {
	bio := C.BIO_new(C.BIO_s_mem())
	if bio == nil {
		return "", GetErrors()
	}
	defer C.BIO_free(bio)
	if C.pem_write_bio_pubkey(bio, pk.pkey) != 1 {
		return "", GetErrors()
	}
	var p *C.char
	pubKeyLen := C.BIO_get_mem_data_func(bio, &p)
	if pubKeyLen <= 0 {
		return "", GetErrors()
	}
	return C.GoString(p)[:pubKeyLen], nil
}

// Sign sign with private key
func (sk *PrivateKey) Sign(alg string, dgst []byte) ([]byte, error) {
	calg := C.CString(alg)
	defer C.free(unsafe.Pointer(calg))
	var siglen C.size_t
	sig := C.sk_sign(sk.pkey, calg, (*C.uchar)(&dgst[0]),
		C.size_t(len(dgst)), &siglen, nil)
	if sig == nil {
		C.ERR_print_errors_fp(C.stderr)
		return nil, GetErrors()
	}
	defer C.free(unsafe.Pointer(sig))
	return C.GoBytes(unsafe.Pointer(sig), C.int(siglen)), nil
}

// Verify verify with public key
func (pk *PublicKey) Verify(alg string, dgst, sig []byte) error {
	calg := C.CString(alg)
	defer C.free(unsafe.Pointer(calg))

	if C.pk_verify(pk.pkey, calg, (*C.uchar)(&dgst[0]), C.size_t(len(dgst)),
		(*C.uchar)(&sig[0]), C.size_t(len(sig)), nil) != 1 {
		C.ERR_print_errors_fp(C.stderr)
		return GetErrors()
	}
	return nil
}

// ComputeSM2IDDigest compute id digest for SM2 by public key
func (pk *PublicKey) ComputeSM2IDDigest(id string) ([]byte, error) {
	if C.EVP_PKEY_EC != C.EVP_PKEY_id(pk.pkey) {
		return nil, errors.New("invalid public key type")
	}
	cid := C.CString(id)
	defer C.free(unsafe.Pointer(cid))
	outbuf := make([]byte, sm2DigestLen*2)
	outlen := C.size_t(len(outbuf))
	if C.SM2_compute_id_digest(C.EVP_sm3(), cid, C.size_t(len(id)),
		(*C.uchar)(&outbuf[0]), &outlen, C.EVP_PKEY_get0_EC_KEY(pk.pkey)) != 1 {
		return nil, GetErrors()
	}
	return outbuf[:sm2DigestLen], nil
}

// ComputeSM2IDDigest compute id digest for SM2 by private key
func (sk *PrivateKey) ComputeSM2IDDigest(id string) ([]byte, error) {
	if C.EVP_PKEY_EC != C.EVP_PKEY_id(sk.pkey) {
		return nil, errors.New("invalid public key type")
	}
	cid := C.CString(id)
	defer C.free(unsafe.Pointer(cid))
	outbuf := make([]byte, 64)
	outlen := C.size_t(len(outbuf))
	if C.SM2_compute_id_digest(C.EVP_sm3(), cid, C.size_t(len(id)),
		(*C.uchar)(&outbuf[0]), &outlen, C.EVP_PKEY_get0_EC_KEY(sk.pkey)) != 1 {
		return nil, GetErrors()
	}
	return outbuf[:sm2DigestLen], nil
}

// GetDer get der format private key
func (sk *PrivateKey) GetDer() ([]byte, error) {
	var der *C.uchar
	var length C.int
	defer func() {
		if der != nil {
			C.free(unsafe.Pointer(der))
		}
	}()
	length = C.i2d_PrivateKey(sk.pkey, &der)
	if length < 1 {
		return nil, errors.New("get der failed for public key")
	}
	goDer := C.GoBytes(unsafe.Pointer(der), length)
	return goDer, nil
}
