/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package action

import (
	"context"
	"strconv"
	"sync"
	"time"

	"github.com/pkg/errors"
	"google.golang.org/grpc"
	"google.golang.org/grpc/connectivity"
	"google.golang.org/grpc/credentials"
	"google.golang.org/grpc/keepalive"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/sdk/crypto"
	"huawei.com/huaweichain/sdk/utils"
)

var log = logger.GetModuleLogger("go sdk", "action")

// Config is the definition of action config.
type Config struct {
	hostInfo       HostInfo
	cpt            crypto.Crypto
	tlsEnable      bool
	transportCreds credentials.TransportCredentials
}

// HostInfo is the definition of host info.
type HostInfo struct {
	HostOverride string
	Host         string
	Port         int
}

// NewConfig is used to create an config instance.
func NewConfig(hostInfo HostInfo, cpt crypto.Crypto, tlsEnable bool,
	transportCreds credentials.TransportCredentials) *Config {
	return &Config{
		hostInfo:       hostInfo,
		cpt:            cpt,
		tlsEnable:      tlsEnable,
		transportCreds: transportCreds,
	}
}

type action struct {
	config *Config
	conn   *grpc.ClientConn
	connMu sync.Mutex
}

func (a *action) resetConn() error {
	_, err := a.GetReadyConn()
	if err != nil {
		return err
	}
	return nil
}

func (a *action) newClientConn() error {
	ipAddr := a.config.hostInfo.Host + ":" + strconv.Itoa(a.config.hostInfo.Port)
	options := a.getOpts()
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	cc, err := grpc.DialContext(ctx, ipAddr, options...)
	if err != nil {
		return errors.WithMessage(err, "grpc dial context error")
	}
	go keep(cc)
	a.conn = cc
	return nil
}

const recoveryTimeout = 30 * time.Second

func (a *action) GetReadyConn() (*grpc.ClientConn, error) {
	if a.conn == nil {
		a.connMu.Lock()
		if a.conn == nil {
			if err := a.newClientConn(); err != nil {
				a.connMu.Unlock()
				return nil, errors.WithMessage(err, "get client connection error")
			}
		}
		a.connMu.Unlock()
	}
	state := a.conn.GetState()
	if state == connectivity.Ready {
		return a.conn, nil
	}
	if state == connectivity.Idle {
		a.conn.Connect()
	}

	ctx, cancel := context.WithTimeout(context.Background(), recoveryTimeout)
	defer cancel()

	for ; state != connectivity.Ready; state = a.conn.GetState() {
		a.conn.ResetConnectBackoff()
		if !a.conn.WaitForStateChange(ctx, state) {
			return nil, errors.Wrapf(ctx.Err(), "failed to wait for connection ready, state[%v]", state)
		}
	}
	return a.conn, nil
}

func keep(cc *grpc.ClientConn) {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	for {
		state := cc.GetState()
		switch state {
		case connectivity.Idle:
			cc.Connect()
		case connectivity.Shutdown:
			return
		default:
			cc.ResetConnectBackoff()
			if !cc.WaitForStateChange(ctx, state) {
				return
			}
		}
	}
}

func (a *action) Close() error {
	return a.conn.Close()
}

func (a *action) getOpts() []grpc.DialOption {
	var opts []grpc.DialOption
	opts = append(opts, grpc.WithBlock())
	if a.config.tlsEnable {
		opts = append(opts, grpc.WithTransportCredentials(a.config.transportCreds))
	} else {
		opts = append(opts, grpc.WithInsecure())
		log.Infof("default insecure for go sdk.....")
	}
	opts = append(opts, grpc.WithKeepaliveParams(keepalive.ClientParameters{
		Time:                10 * time.Second,
		Timeout:             2 * time.Second,
		PermitWithoutStream: true,
	}))
	return opts
}
