/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package event

import (
	"context"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
)

// ContractEventService is the definition of block event service.
type ContractEventService struct {
	chainID string
	client  nodeservice.EventServiceClient
	ctx     context.Context
	cancel  context.CancelFunc
}

// NewContractEventService is used to create an instance of block event service.
func NewContractEventService(client nodeservice.EventServiceClient, chainID string) *ContractEventService {
	ctx, cancel := context.WithCancel(context.Background())
	return &ContractEventService{
		chainID: chainID,
		client:  client,
		ctx:     ctx,
		cancel:  cancel,
	}
}

// RegisterBlockEvent is used to register block event from latest block.
func (c *ContractEventService) RegisterBlockEvent() (*ContractIterator, error) {
	payload := []byte(c.chainID)
	rawMsg := &common.RawMessage{
		Payload: payload,
	}

	client, err := c.client.RegisterContractEvent(c.ctx, rawMsg)
	if err != nil {
		return nil, err
	}
	return &ContractIterator{
		client: client,
		ctx:    c.ctx,
	}, nil
}

// ContractIterator is the definition of contract iterator.
type ContractIterator struct {
	client nodeservice.EventService_RegisterContractEventClient
	ctx    context.Context
}

// Next get next contract event
func (c ContractIterator) Next() (*common.ContractEvent, error) {
	rawMsg, err := c.client.Recv()
	if err != nil {
		return nil, errors.WithMessage(err, "recv contract event error")
	}
	select {
	case <-c.ctx.Done():
		return nil, errors.New("grpc stream has closed")
	default:
		var response common.Response
		if err = proto.Unmarshal(rawMsg.Payload, &response); err != nil {
			return nil, errors.WithMessage(err, "unmarshal Response error")
		}
		if response.Status != common.SUCCESS {
			return nil, errors.Errorf("response status is: %v", response.Status)
		}
		event := common.ContractEvent{}
		if err = proto.Unmarshal(response.Payload, &event); err != nil {
			return nil, errors.WithMessage(err, "unmarshal contract event error")
		}
		return &event, nil
	}
}
