/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package event

import (
	"context"
	"strings"
	"time"

	"google.golang.org/grpc/codes"
	rpcStatus "google.golang.org/grpc/status"

	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/rawmessage"
)

const (
	rpcErrDesc = "node is still under initialization"
)

type listenerParam struct {
	ctx        context.Context
	cancel     context.CancelFunc
	chainID    string
	ch         chan msg
	msgBuilder rawmessage.MsgBuilder
	client     nodeservice.EventServiceClient
}

type msg interface{}

type registerMsg struct {
	id         string
	statusCh   chan status
	dispatchCh chan *nodeservice.TxEventRes
	resultCh   chan *common.TxResult
}

type unregisterMsg struct {
	id       string
	statusCh chan status
}
type txEventResMsg struct {
	txEventRes *nodeservice.TxEventRes
}

type txResultMsg struct {
	txResult *common.TxResult
}

type sendMsg struct {
	ch      chan status
	txEvent *nodeservice.TxEvent
}

type eventListener interface {
	init() error
	registerTx(txHash []byte) (chan *common.TxResult, error)
}

func listen(recv func() (*common.RawMessage, error), processor func(rawMsg *common.RawMessage), reInit func() error) {
	for {
		responseMsg, err := recv()
		if err == nil {
			processor(responseMsg)
			continue
		}
		rpcErr, ok := rpcStatus.FromError(err)
		if !ok {
			log.Errorf("recv message error, please recover service manually, error: %v", err)
			return
		}
		// client close stream actively
		if rpcErr.Code() == codes.Canceled {
			return
		}
		if rpcErr.Code() != codes.Unavailable {
			if rpcErr.Code() != codes.Unknown || strings.TrimSpace(rpcErr.Message()) != rpcErrDesc {
				log.Errorf("recv error: %v, please recover service manually", rpcErr)
				return
			}
		}
		if err := reInit(); err != nil {
			log.Errorf("re init failed: %v, please recover service manually", err)
		}
		return
	}
}

func reInit(registerEvent func() error) error {
	for {
		log.Infof("re register event after %d second", reconnectIntervalSecond)
		time.Sleep(reconnectIntervalSecond * time.Second)
		log.Infof("re register event......")
		err := registerEvent()
		if err == nil {
			break
		}

		log.Errorf("re register event error: %v", err)
		rpcErr, ok := rpcStatus.FromError(err)
		if !ok {
			return err
		}
		// client close stream actively
		if rpcErr.Code() == codes.Canceled {
			return nil
		}
		if rpcErr.Code() != codes.Unavailable {
			if rpcErr.Code() != codes.Unknown || strings.TrimSpace(rpcErr.Message()) != rpcErrDesc {
				return err
			}
		}
	}
	log.Info("re register event success")
	return nil
}
