/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package event

import (
	"context"
	"encoding/hex"
	"sync"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/rawmessage"
)

// SourceType is the data source type of tx event service.
type SourceType int32

const (
	buffSize = 3000
	// Block is the source type of block event.
	Block SourceType = 0
	// Tx is the source type of tx event.
	Tx SourceType = 1

	reconnectIntervalSecond = 3
)

// TxEventService 支持Block和Tx两种不同的Event来源
type TxEventService struct {
	txMap     map[string]chan *common.TxResult
	chainID   string
	ch        chan msg
	listener  eventListener
	ctx       context.Context
	cancel    context.CancelFunc
	isInit    bool
	initLock  sync.Mutex
	holdCount int32
	clean     func()
}

// NewTxEventService is used to create an instance of tx event service, default with tx event data source.
func NewTxEventService(client nodeservice.EventServiceClient, chainID string,
	msgBuilder rawmessage.MsgBuilder, clean func()) *TxEventService {
	return NewTxEventServiceWithSourceType(client, chainID, msgBuilder, Block, clean)
}

// NewTxEventServiceWithSourceType is used to create an instance of tx event service with specified event data source.
func NewTxEventServiceWithSourceType(client nodeservice.EventServiceClient, chainID string,
	msgBuilder rawmessage.MsgBuilder, source SourceType, clean func()) *TxEventService {
	ctx, cancel := context.WithCancel(context.Background())
	ch := make(chan msg, buffSize)
	params := &listenerParam{
		ctx:        ctx,
		cancel:     cancel,
		chainID:    chainID,
		ch:         ch,
		msgBuilder: msgBuilder,
		client:     client,
	}
	var listener eventListener
	if source == Block {
		listener = newBlockListener(params)
	} else {
		listener = newTxListener(params)
	}
	return &TxEventService{
		chainID:   chainID,
		ch:        ch,
		txMap:     make(map[string]chan *common.TxResult),
		listener:  listener,
		ctx:       ctx,
		cancel:    cancel,
		holdCount: 1,
		clean:     clean,
	}
}

func (s *TxEventService) init() error {
	if s.isInit {
		return nil
	}
	s.initLock.Lock()
	defer s.initLock.Unlock()
	if s.isInit {
		return nil
	}

	err := s.listener.init()
	if err != nil {
		return errors.WithMessage(err, "listener init error")
	}

	go s.dispatch()
	s.isInit = true
	return nil
}

// RegisterTx is used to register transaction.
func (s *TxEventService) RegisterTx(txHash []byte) (chan *common.TxResult, error) {
	if err := s.init(); err != nil {
		return nil, errors.WithMessage(err, "tx event service init error")
	}
	select {
	case <-s.ctx.Done():
		return nil, errors.New("tx event service has already closed")
	default:
		return s.listener.registerTx(txHash)
	}
}

// TryHold is used to hold tx event service.
func (s *TxEventService) TryHold() bool {
	return tryHold(&s.holdCount)
}

// Close is used to close tx event service.
func (s *TxEventService) Close() {
	count := close(&s.holdCount)
	log.Debugf("close tx event service, count: %v", count)
	if count == 0 {
		s.clean()
		s.cancel()
	}
}

func (s *TxEventService) dispatch() {
	for {
		select {
		case m := <-s.ch:
			s.processMsg(m)
		case <-s.ctx.Done():
			return
		}
	}
}

func (s *TxEventService) processMsg(m msg) {
	switch u := m.(type) {
	case *registerMsg:
		s.txMap[u.id] = u.resultCh
		u.statusCh <- success
	case *txResultMsg:
		txID := hex.EncodeToString(u.txResult.TxHash)
		ch, ok := s.txMap[txID]
		if !ok {
			return
		}
		if ch == nil {
			log.Error("tx result chan is nil")
			delete(s.txMap, txID)
			return
		}
		ch <- u.txResult
		delete(s.txMap, txID)
	default:
	}
}
