/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package action

import (
	"context"
	"sync"
	"time"

	"github.com/pkg/errors"
	"google.golang.org/grpc/grpclog"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/utils"
)

// QueryAction is the action for query operations.
type QueryAction struct {
	action
	chainServiceClient nodeservice.ChainServiceClient
	circleClient       nodeservice.CircleClient
	voteManagerClient  nodeservice.VoteManagerClient
	clientLck          sync.Mutex
}

// NewQueryAction is used to create query action instance with action config.
func NewQueryAction(config *Config) *QueryAction {
	action := &QueryAction{action: action{config: config}}
	return action
}

// GetListenEvents is used to listen events of contract.
func (action *QueryAction) GetListenEvents(rawMsg *common.RawMessage) (*StreamListenEventsIterator, error) {
	createStream := func(client nodeservice.ChainServiceClient,
		rawMsg *common.RawMessage) (interface{}, error) {
		return createListenEventsStream(client, rawMsg)
	}

	stream, err := action.getStreamIterator(rawMsg, createStream)
	if err != nil {
		return nil, errors.WithMessage(err, "get stream iterator error")
	}
	streamResultIterator, ok := stream.(*StreamListenEventsIterator)
	if !ok {
		return nil, errors.Errorf("type assert error: expected type: *StreamListenEventsIterator, but got: %T",
			stream)
	}
	return streamResultIterator, nil
}

func createListenEventsStream(client nodeservice.ChainServiceClient,
	rawMsg *common.RawMessage) (*StreamListenEventsIterator, error) {
	stream, err := client.GetListenEvents(context.Background())
	if err != nil {
		return nil, errors.WithMessage(err, "GetStateOfChain error")
	}
	ctx, cancel := context.WithCancel(context.Background())
	streamResultIterator := &StreamListenEventsIterator{client: stream, ctx: ctx, cancel: cancel}
	err = stream.Send(rawMsg)
	if err != nil {
		cancel()
		grpclog.Errorf("Stream send err:%s", err.Error())
	}
	return streamResultIterator, nil
}

// StreamListenEventsIterator is the definition of listen events result iterator.
type StreamListenEventsIterator struct {
	data   []byte
	client nodeservice.ChainService_GetListenEventsClient
	ctx    context.Context
	cancel context.CancelFunc
	err    error
}

// Next is used to get the next element of iterator.
func (itr *StreamListenEventsIterator) Next() bool {
	streamMsg, err := itr.client.Recv()
	itr.err = err
	if err != nil {
		itr.err = err
		return false
	}
	select {
	case <-itr.ctx.Done():
		itr.err = errors.New("grpc stream has closed")
		return false
	default:
		itr.data = streamMsg.Payload
		return true
	}
}

// Value is used to get the next element's value of iterator.
func (itr *StreamListenEventsIterator) Value() []byte {
	return itr.data
}

// Error is used to get error.
func (itr *StreamListenEventsIterator) Error() error {
	return itr.err
}

// Cancel is used to cancel this data chan.
func (itr *StreamListenEventsIterator) Cancel() {
	err := itr.client.CloseSend()
	if err != nil {
		itr.cancel()
		return
	}
	itr.err = err
}

// GetStateOfChain is used to send get state data of chain.
func (action *QueryAction) GetStateOfChain(rawMsg *common.RawMessage) (*StreamChainStateIterator, error) {
	createStream := func(client nodeservice.ChainServiceClient,
		rawMsg *common.RawMessage) (interface{}, error) {
		return createChainStateStream(client, rawMsg)
	}

	stream, err := action.getStreamIterator(rawMsg, createStream)
	if err != nil {
		return nil, errors.WithMessage(err, "get stream iterator error")
	}
	streamResultIterator, ok := stream.(*StreamChainStateIterator)
	if !ok {
		return nil, errors.Errorf("type assert error: expected type: *StreamChainStateIterator, but got: %T",
			stream)
	}
	return streamResultIterator, nil
}

func (action *QueryAction) getStreamIterator(rawMsg *common.RawMessage,
	createStream func(client nodeservice.ChainServiceClient,
		rawMsg *common.RawMessage) (interface{}, error)) (interface{}, error) {
	client, err := action.getChainServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get chain service client error")
	}
	for i := 0; i < maxStreamConnTryTimes; i++ {
		streamIterator, err := createStream(client, rawMsg)
		if err != nil {
			grpclog.Errorf("create stream err:%s, the try time is %d", err.Error(), i)
			time.Sleep(time.Second * time.Duration(i+1))
			continue
		}
		return streamIterator, nil
	}
	return nil, errors.New("connect error")
}

func createChainStateStream(client nodeservice.ChainServiceClient,
	rawMsg *common.RawMessage) (*StreamChainStateIterator, error) {
	stream, err := client.GetStateOfChain(context.Background())
	if err != nil {
		return nil, errors.WithMessage(err, "GetStateOfChain error")
	}
	ctx, cancel := context.WithCancel(context.Background())
	streamResultIterator := &StreamChainStateIterator{client: stream, ctx: ctx, cancel: cancel}
	err = stream.Send(rawMsg)
	if err != nil {
		cancel()
		grpclog.Errorf("Stream send err:%s", err.Error())
	}
	return streamResultIterator, nil
}

// StreamChainStateIterator is the definition of chain state result iterator.
type StreamChainStateIterator struct {
	data   *nodeservice.ChainState
	client nodeservice.ChainService_GetStateOfChainClient
	ctx    context.Context
	cancel context.CancelFunc
	err    error
}

// Next is used to get the next element of iterator.
func (itr *StreamChainStateIterator) Next() bool {
	streamMsg, err := itr.client.Recv()
	itr.err = err
	if err != nil {
		itr.err = err
		return false
	}
	select {
	case <-itr.ctx.Done():
		itr.err = errors.New("grpc stream has closed")
		return false
	default:
		data := nodeservice.ChainState{}
		err = proto.Unmarshal(streamMsg.Payload, &data)
		if err != nil {
			return false
		}
		itr.data = &data
		return true
	}
}

// Key is used to get the next element's key of iterator.
func (itr *StreamChainStateIterator) Key() string {
	return itr.data.Key
}

// Value is used to get the next element's value of iterator.
func (itr *StreamChainStateIterator) Value() []byte {
	return itr.data.Value
}

// Contract is used to get the next element's value of iterator.
func (itr *StreamChainStateIterator) Contract() string {
	return itr.data.ContractName
}

// Error is used to get error.
func (itr *StreamChainStateIterator) Error() error {
	return itr.err
}

// Cancel is used to cancel this data chan.
func (itr *StreamChainStateIterator) Cancel() {
	err := itr.client.CloseSend()
	if err != nil {
		itr.cancel()
		return
	}
	itr.err = err
}

// GetLatestChainState is used to send get latest chain request raw message by grpc.
func (action *QueryAction) GetLatestChainState(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getChainServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get chain service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.GetLatestChainState(ctx, rawMsg)
}

// GetBlockByNum is used to send get block by number request raw message by grpc.
func (action *QueryAction) GetBlockByNum(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getChainServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get chain service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.GetBlockByNum(ctx, rawMsg)
}

// QueryCircleByName is used to send query circle by name request raw message by grpc.
func (action *QueryAction) QueryCircleByName(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getCircleServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get circle service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.QueryCircleConfig(ctx, rawMsg)
}

// QueryCircles is used to send query circles raw message by grpc.
func (action *QueryAction) QueryCircles(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getCircleServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get circle service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.QueryChainCircles(ctx, rawMsg)
}

// GetBlockAndResultByNum is used to send get block and block result by number request raw message by grpc.
func (action *QueryAction) GetBlockAndResultByNum(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getChainServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get chain service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.GetBlockAndResultByNum(ctx, rawMsg)
}

// GetBlockByTxHash is used to send get block by transaction id request raw message by grpc.
func (action *QueryAction) GetBlockByTxHash(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getChainServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get chain service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.GetBlockByTxHash(ctx, rawMsg)
}

// GetTxByHash is used to send get transaction by transaction id request raw message by grpc.
func (action *QueryAction) GetTxByHash(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	if rawMsg == nil {
		return nil, errors.New("raw message is nil")
	}
	client, err := action.getChainServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get chain service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.GetTxByHash(ctx, rawMsg)
}

// GetTxResultByTxHash is used to send get tx result by transaction id request raw message by grpc.
func (action *QueryAction) GetTxResultByTxHash(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	if rawMsg == nil {
		return nil, errors.New("raw message is nil")
	}
	client, err := action.getChainServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get chain service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.GetTxResultByTxHash(ctx, rawMsg)
}

// GetContractInfo is used to send get contract info request raw message by grpc.
func (action *QueryAction) GetContractInfo(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getChainServiceClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get chain service client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.GetContractInfo(ctx, rawMsg)
}

// GetVote is used to send get vote request raw message by grpc.
func (action *QueryAction) GetVote(rawMsg *common.RawMessage) (*common.RawMessage, error) {
	client, err := action.getVoteManagerClient()
	if err != nil {
		return nil, errors.WithMessage(err, "get vote manager client error")
	}
	ctx, cancel := context.WithTimeout(context.Background(), utils.GetTimeout()*time.Second)
	defer cancel()
	return client.Query(ctx, rawMsg)
}

func (action *QueryAction) getChainServiceClient() (nodeservice.ChainServiceClient, error) {
	if err := action.resetConn(); err != nil {
		return nil, errors.WithMessage(err, "reset clients error")
	}
	if action.chainServiceClient == nil {
		action.newClients()
	}
	return action.chainServiceClient, nil
}

func (action *QueryAction) getCircleServiceClient() (nodeservice.CircleClient, error) {
	if err := action.resetConn(); err != nil {
		return nil, errors.WithMessage(err, "reset clients error")
	}
	if action.circleClient == nil {
		action.newClients()
	}
	return action.circleClient, nil
}

func (action *QueryAction) getVoteManagerClient() (nodeservice.VoteManagerClient, error) {
	if err := action.resetConn(); err != nil {
		return nil, errors.WithMessage(err, "reset clients error")
	}
	if action.voteManagerClient == nil {
		action.newClients()
	}
	return action.voteManagerClient, nil
}

func (action *QueryAction) newClients() {
	action.clientLck.Lock()
	defer action.clientLck.Unlock()
	if action.chainServiceClient == nil || action.voteManagerClient == nil || action.circleClient == nil {
		action.chainServiceClient = nodeservice.NewChainServiceClient(action.conn)
		action.voteManagerClient = nodeservice.NewVoteManagerClient(action.conn)
		action.circleClient = nodeservice.NewCircleClient(action.conn)
	}
}
