/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

// Package flow is the implementation of flow sdk for go client.
package flow

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/config"
	"huawei.com/huaweichain/sdk/node"
)

// ChainFlow is used to do the chain related operation
// which includes create, delete and query
type ChainFlow struct {
	*ObserverFlow
	genesis    string // chain config path
	entrypoint *common.Entrypoint
	decrypt    func(bytes []byte) ([]byte, error)
}

// newChainFlow is used to return a new ChainFlow
func newChainFlow(stub *stub) *ChainFlow {
	return &ChainFlow{
		ObserverFlow: &ObserverFlow{
			Flow: &Flow{
				stub: stub,
			},
			Node: &node.WNode{},
		},
	}
}

// SetTimeout is used to set timeout for sdk
func (c *ChainFlow) SetTimeout(timeout int64) *ChainFlow {
	c.Flow.timeout = timeout
	return c
}

// SetDecrypt is used to set decrypt for sdk
func (c *ChainFlow) SetDecrypt(decrypt func(bytes []byte) ([]byte, error)) *ChainFlow {
	c.decrypt = decrypt
	return c
}

// SetChain is used to set ChainId for sdk
func (c *ChainFlow) SetChain(chain string) *ChainFlow {
	c.ObserverFlow.SetChain(chain)
	return c
}

// SetNode is used to set Node for sdk
func (c *ChainFlow) SetNode(node string) *ChainFlow {
	_, err := c.ObserverFlow.SetNodeWithErr(node)
	if err != nil {
		log.Fatal(err.Error())
	}
	return c
}

// SetEntrypoint is used to set EntryPoint for sdk
func (c *ChainFlow) SetEntrypoint(zone string, master bool, seeds []string) *ChainFlow {
	e := &common.Entrypoint{
		ZoneId:        zone,
		Coordinator:   false,
		InitialMaster: master,
	}
	if master {
		e.Linkers = seeds
	} else {
		e.Seeds = seeds
	}
	c.entrypoint = e
	return c
}

// SetGenesis is used to set genesis path for sdk
func (c *ChainFlow) SetGenesis(genesis string) *ChainFlow {
	c.genesis = genesis
	return c
}

// CreateChain is used to create chain for sdk
func (c *ChainFlow) CreateChain() error {
	// read chain config from file
	c.stub.gatewayClient.SetTimeout(c.timeout)
	genesisConfig, err := config.NewGenesisConfig(c.genesis)
	if err != nil {
		return errors.WithMessage(err, "new genesis config error")
	}

	genesisBlock, err := c.stub.GetChainRawMessage().BuildGenesisBlockWithGenesisConfig(c.chain, genesisConfig)
	if err != nil {
		return errors.WithMessage(err, "build genesis block error")
	}

	genesisBlockBytes, err := genesisBlock.Marshal()
	if err != nil {
		return errors.WithMessage(err, "marshal genesis block error")
	}

	// construct raw message
	rawMsg, err := c.stub.GetChainRawMessage().BuildJoinMsgWithEntrypoint(genesisBlockBytes, c.entrypoint)

	// send message
	if c.Node == nil {
		return errors.WithMessage(err, "need to set node")
	}
	_, err = c.Node.ChainAction.JoinChain(rawMsg)
	if err != nil {
		return errors.WithMessage(err, "chain action join chain error")
	}

	return nil
}

// JoinChain is used to join chain for sdk
func (c *ChainFlow) JoinChain(existedNode string) error {
	c.stub.gatewayClient.SetTimeout(c.timeout)
	genesisBlock, err := c.getGenesisBlockResponse(existedNode)
	chainInfo, infoErr := c.QueryChain(existedNode, c.chain)
	if infoErr != nil {
		return errors.WithMessage(infoErr, "get config info error")
	}
	if c.entrypoint == nil {
		return errors.WithMessage(err, "need to set entryPoint config")
	}
	rawMsg, err := c.stub.GetChainRawMessage().BuildJoinMsgWithLatestConf(genesisBlock.Payload,
		chainInfo, c.entrypoint)
	if c.Node == nil {
		return errors.WithMessage(err, "need to set node")
	}
	responseMsg, err := c.Node.ChainAction.JoinChain(rawMsg)
	if err != nil {
		return errors.WithMessage(err, "chain action join chain error")
	}
	if err := proto.Unmarshal(responseMsg.Payload, genesisBlock); err != nil {
		return errors.WithMessage(err, "unmarshal response error")
	}
	if genesisBlock.Status != common.SUCCESS {
		return errors.Errorf("response error: status: %v, info: %v", genesisBlock.Status.String(), genesisBlock.StatusInfo)
	}

	return nil
}

func (c *ChainFlow) getGenesisBlockResponse(existedNode string) (*common.Response, error) {
	rawMsg, err := c.stub.GetQueryRawMessage().BuildBlockRawMessage(c.chain, 0)
	if err != nil {
		return nil, errors.WithMessage(err, "fail in build block raw message")
	}
	nodeMap := c.stub.gatewayClient.Nodes
	node, ok := nodeMap[existedNode] //
	if !ok {
		return nil, errors.Errorf("exist node not exist： %v", existedNode)
	}
	response, err := node.QueryAction.GetBlockByNum(rawMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "query action query block by number error")
	}
	genesisBlock := &common.Response{}
	if err = proto.Unmarshal(response.Payload, genesisBlock); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response error: %v")
	}
	return genesisBlock, nil
}

func (c *ChainFlow) getChainInfo(chain string, node *node.WNode) (*common.ConfigInfo, error) {
	c.stub.gatewayClient.SetTimeout(c.timeout)
	req, err := c.stub.GetChainRawMessage().BuildQueryChainRawMessage(chain)
	if err != nil {
		return nil, errors.WithMessage(err, "build query single chain raw message error")
	}
	rawRes, err := node.ChainAction.QueryChain(req)
	if err != nil {
		return nil, errors.WithMessage(err, "chain action query single chain error")
	}
	var res common.Response
	if err := proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response error")
	}
	if res.Status != common.SUCCESS {
		return nil, errors.Errorf("response error: status: %v, info: %v", res.Status.String(), res.StatusInfo)
	}
	var queryInfoResponse nodeservice.QueryInfoResponse
	if err := proto.Unmarshal(res.Payload, &queryInfoResponse); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response error")
	}
	info := queryInfoResponse.ChainInfo
	if info == nil {
		return nil, errors.Errorf("empty chain information from %s for chain %s", c.Node.ID, chain)
	}
	return info.Config, nil
}

// QueryChain is used to query chain for sdk
func (c *ChainFlow) QueryChain(node string, chain string) (*common.ConfigInfo, error) {
	wNode, err := c.stub.GetWNode(node)
	if err != nil {
		return nil, err
	}
	return c.getChainInfo(chain, wNode)
}

// Exist returns true if the node has been joined chain.
func (c *ChainFlow) Exist(node string) error {
	if _, err := c.QueryChain(node, c.chain); err != nil {
		return errors.WithMessagef(err, "failed to find chain in node %s", node)
	}
	return nil
}

// ChainConf is used to query chain for sdk
func (c *ChainFlow) ChainConf() (*common.ConfigInfo, error) {
	return c.getChainInfo(c.chain, c.Node)
}

// DeleteChain is used to delete chain for sdk
func (c *ChainFlow) DeleteChain() error {
	c.stub.gatewayClient.SetTimeout(c.timeout)
	rawMsg, err := c.stub.GetChainRawMessage().BuildQuitChainRawMessage(c.chain)
	if err != nil {
		return errors.WithMessage(err, "build quit chain raw message error")
	}
	if c.Node == nil {
		return errors.WithMessage(err, "need to set node")
	}
	_, err = c.Node.ChainAction.QuitChain(rawMsg)
	if err != nil {
		return errors.WithMessage(err, "chain action quit chain error")
	}
	return nil
}

// DeleteChainForAllNodes is used to delete chain for all nodes for sdk
func (c *ChainFlow) DeleteChainForAllNodes() error {
	c.stub.gatewayClient.SetTimeout(c.timeout)
	rawMsg, err := c.stub.GetChainRawMessage().BuildQuitChainRawMessage(c.chain)
	if err != nil {
		return errors.WithMessage(err, "build quit chain raw message error")
	}
	nodeMap := c.stub.gatewayClient.Nodes
	for _, node := range nodeMap {
		_, err := node.ChainAction.QuitChain(rawMsg)
		if err != nil {
			return errors.WithMessage(err, "chain action quit chain error")
		}
	}
	return nil
}

// QueryAllChain is used to query all chains for sdk
func (c *ChainFlow) QueryAllChain() (*common.RawMessage, error) {
	c.stub.gatewayClient.SetTimeout(c.timeout)
	rawMsg, err := c.stub.GetChainRawMessage().BuildQueryAllChainRawMessage()
	if err != nil {
		return nil, errors.WithMessage(err, "build query all chains raw message error")
	}
	if c.Node == nil {
		return nil, errors.WithMessage(err, "need to set node")
	}
	responseMsg, err := c.Node.ChainAction.QueryAllChains(rawMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "chain action query all chains error")
	}
	response := &common.Response{}
	if err := proto.Unmarshal(responseMsg.Payload, response); err != nil {
		return nil, errors.WithMessage(err, "unmarshal response error")
	}
	if response.Status != common.SUCCESS {
		return nil, errors.Errorf("response error: status: %v, info: %v", response.Status.String(), response.StatusInfo)
	}
	queryAllInfo := &nodeservice.QueryAllInfoResponse{}
	if err := proto.Unmarshal(response.Payload, queryAllInfo); err != nil {
		return nil, errors.WithMessage(err, "unmarshal query all info error")
	}
	return responseMsg, nil
}
