/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package flow

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/proto/schema"
	"huawei.com/huaweichain/sdk/node"
	"huawei.com/huaweichain/sdk/rawmessage"
)

// ConfigFlow is used to do the config related operation
type ConfigFlow struct {
	*consenterFlow
}

func newConfigFlow(stub *stub, node *node.WNode) *ConfigFlow {
	configFlowInstance := ConfigFlow{
		consenterFlow: &consenterFlow{
			performerFlow: &performerFlow{
				ObserverFlow: &ObserverFlow{
					Flow: &Flow{
						stub: stub,
					},
					Node: node,
				},
			},
		},
	}
	return &configFlowInstance
}

// QueryChainVoteUpdate is used to query chain vote process
func (c *ConfigFlow) QueryChainVoteUpdate(chainID string, subject string) (*common.RawMessage, error) {
	c.stub.gatewayClient.SetTimeout(c.timeout)
	rawMsg, err := c.stub.GetQueryRawMessage().BuildQueryChainUpdateVoteRawMessage(chainID, subject)
	if err != nil {
		return nil, errors.WithMessage(err, "build query chain update vote raw message error")
	}
	if c.Node == nil {
		return nil, errors.WithMessage(err, "need to set node")
	}
	responseRaw, err := c.Node.QueryAction.GetVote(rawMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "chain action query get vote error")
	}
	response := &common.Response{}
	if err = proto.Unmarshal(responseRaw.Payload, response); err != nil {
		return nil, errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if response.Status != common.SUCCESS {
		return nil, errors.Errorf("response error: status: %v, info: %v", response.Status.String(), response.StatusInfo)
	}
	voteQueryResponse := &nodeservice.VoteQueryResponse{}
	if err := proto.Unmarshal(response.Payload, voteQueryResponse); err != nil {
		return nil, errors.WithMessage(err, "unmarshal vote query response error")
	}
	return responseRaw, nil
}

// SetNode is used to set node
func (c *ConfigFlow) SetNode(node string) *ConfigFlow {
	c.Node = c.stub.gatewayClient.Nodes[node]
	return c
}

// SetChain sets chain id.
func (c *ConfigFlow) SetChain(chain string) *ConfigFlow {
	c.consenterFlow.SetChain(chain)
	return c
}

// SetTimeout sets request timeout.
func (c *ConfigFlow) SetTimeout(timeout int64) *ConfigFlow {
	c.timeout = timeout
	return c
}

// UpdatePolicy is used to update policy
func (c *ConfigFlow) UpdatePolicy(chainID string, policy string) error {
	c.stub.gatewayClient.SetTimeout(c.timeout)
	rawMsg, err := c.stub.GetConfigRawMessage().UpdateConfig.BuildUpdateConfPolicyRawMessage(chainID, policy)
	if err != nil {
		return errors.WithMessage(err, "build update config policy raw message error")
	}
	err = c.SendTx(rawMsg)
	if err != nil {
		return errors.WithMessage(err, "fail to process sending tx ")
	}
	return nil
}

// UpdateLifecycle is used to update lifecycle
func (c *ConfigFlow) UpdateLifecycle(chainID string, policy string) error {
	c.stub.gatewayClient.SetTimeout(c.timeout)
	rawMsg, err := c.stub.GetConfigRawMessage().UpdateConfig.BuildUpdateLifecycleRawMessage(chainID, policy)
	if err != nil {
		return errors.WithMessage(err, "build update lifecycle raw message error")
	}
	err = c.SendTx(rawMsg)
	if err != nil {
		return errors.WithMessage(err, "fail to process sending tx ")
	}
	return nil
}

// VoteType represents vote transaction type.
type VoteType int

const (
	// ConfigVote represents the vote tx is vote command.
	ConfigVote VoteType = iota
	// ConfigComplete represents the vote tx is complete command.
	ConfigComplete
)

// AddConsenter sends vote transaction to change consensus configuration to add consensus node.
func (c *ConfigFlow) AddConsenter(vote VoteType, consenter ...Consenter) error {
	ccs := make([]*schema.Consenter, len(consenter))
	for i := 0; i < len(consenter); i++ {
		cons := &consenter[i]
		ccs[i] = &schema.Consenter{
			Name:     cons.Name,
			Org:      cons.Org,
			Host:     cons.Addr,
			Port:     uint64(cons.RpcPort),
			ReqPort:  cons.ReqPort,
			RestPort: cons.RestPort,
			ReeCert:  cons.PemCert,
			TeeCert:  cons.PemCert,
		}
	}
	updates := schema.ConsenterUpdates{
		Op:         schema.AddVoter,
		Consenters: ccs,
	}
	return c.consenterUpdate(&updates, convertConsenterVoteHandler(vote))
}

func convertConsenterVoteHandler(vote VoteType) string {
	var handler string
	switch vote {
	case ConfigVote:
		handler = rawmessage.VoteHandler
	case ConfigComplete:
		handler = rawmessage.CompleteHandler
	default:
		log.Panicf("invalid vote type %d", vote)
	}
	return handler
}

func (c *ConfigFlow) consenterUpdate(updates *schema.ConsenterUpdates, handler string) error {
	update := schema.ConsensusUpdate{
		Update: &schema.ConsensusUpdate_ConfChange{
			ConfChange: updates,
		},
	}
	updateBytes, err := proto.Marshal(&update)
	if err != nil {
		return errors.Wrapf(err, "failed to marshal %T", updates)
	}
	set := common.ConfigSet{
		Value: &common.ConfigSet_ConsenterUpdate{
			ConsenterUpdate: updateBytes,
		},
	}
	setBytes, err := proto.Marshal(&set)
	if err != nil {
		return errors.Wrapf(err, "failed to marshal %T", set)
	}
	builder := rawmessage.NewMsgBuilderImpl(c.stub.GetCrypto())
	data := common.VoteTxData{
		Handler: handler,
		Payload: setBytes,
	}
	tx, err := builder.BuildVoteTx(c.chain, &data)
	if err != nil {
		return errors.WithMessage(err, "build add tx error")
	}
	msg, err := builder.GetTxRawMsg(tx)
	if err != nil {
		return err
	}
	return c.SendTx(msg)
}

// RemoveConsenter sends vote transaction to change consensus configuration to remove consensus node.
func (c *ConfigFlow) RemoveConsenter(vote VoteType, consenter ...string) error {
	ccs := make([]*schema.Consenter, len(consenter))
	for i, cons := range consenter {
		ccs[i] = &schema.Consenter{Name: cons}
	}
	updates := schema.ConsenterUpdates{
		Op:         schema.Remove,
		Consenters: ccs,
	}
	return c.consenterUpdate(&updates, convertConsenterVoteHandler(vote))
}

// SetConsenterList is used to set consenter list
func (c *ConfigFlow) SetConsenterList(consenter []string) *ConfigFlow {
	_, err := c.consenterFlow.SetConsenterList(consenter)
	if err != nil {
		log.Fatal(err.Error())
	}
	return c
}
