/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package flow

import (
	"bytes"
	"context"
	"fmt"
	"sort"
	"strconv"
	"sync"
	"time"

	"github.com/pkg/errors"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/schema"
	"huawei.com/huaweichain/sdk/flow/rest"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/sdk/client"
	"huawei.com/huaweichain/sdk/utils"
)

var consLg = logger.GetModuleLogger("sdk.flow.consensus", "config_change")

// Consenter represents consensus node information.
type Consenter struct {
	Name     string
	Org      string
	Addr     string
	RpcPort  uint32 // nolint:golint,stylecheck // ignore
	ReqPort  uint32
	RestPort uint32
	PemCert  []byte
}

// AddConsenter adds multiple nodes to consensus cluster in batch. If any node has not been joined chain,
// try to add it to zone cluster as peer first, then synchronize block progress and transfer to consensus node.
func AddConsenter(client *client.GatewayClient, chain string, consensusNode string, seeds []string,
	consenter ...Consenter) (*Result, error) {
	if len(chain) == 0 {
		return nil, errors.Wrap(ErrEmptyChainID, "failed to add consenter")
	}
	if len(consensusNode) == 0 {
		return nil, errors.New("consensus node should not be empty")
	}
	if _, ok := client.Nodes[consensusNode]; !ok {
		return nil, errors.Errorf("consensus node %s doesn't exist", consensusNode)
	}
	if len(consenter) == 0 {
		return nil, errors.New("empty consenter list which add to consensus cluster")
	}
	for i := range consenter {
		n := &consenter[i]
		if len(n.Name) == 0 {
			return nil, errors.New("consenter name is empty")
		}
		if _, ok := client.Nodes[n.Name]; !ok {
			return nil, errors.Errorf("pending consensus node %s doesn't exist in local network configuration", n.Name)
		}
	}
	sort.Slice(consenter, func(i, j int) bool {
		return consenter[i].Name < consenter[j].Name
	})
	result, childCtx := newResult(context.Background())
	a := addProcessor{
		client:     client,
		chain:      chain,
		existNode:  consensusNode,
		seeds:      seeds,
		consenters: consenter,
		res:        result,
	}
	if err := a.validate(); err != nil {
		return nil, err
	}
	go a.add(childCtx)
	return result, nil
}

func (p *addProcessor) validate() error {
	ns := make([]string, len(p.consenters))
	for i := range p.consenters {
		ns[i] = p.consenters[i].Name
	}
	if err := validateName(ns); err != nil {
		return err
	}
	nodes := make([]Consenter, len(p.consenters))
	copy(nodes, p.consenters)
	sort.Slice(nodes, func(i, j int) bool {
		return bytes.Compare(nodes[i].PemCert, nodes[j].PemCert) <= 0
	})

	for i := range nodes {
		n := &nodes[i]
		if len(n.Addr) == 0 {
			return errors.Errorf("node %s address is empty", n.Name)
		}
		if len(n.PemCert) == 0 {
			return errors.Errorf("node %s pem certificate is empty", n.Name)
		}
		if i == len(nodes)-1 {
			break
		}
		if bytes.Equal(n.PemCert, nodes[i+1].PemCert) {
			return errors.Errorf("duplicate node %s and node %s cert in consenter change list", n.Name, nodes[i+1].Name)
		}
	}
	return nil
}

func validateName(consenters []string) error {
	nodes := make([]string, len(consenters))
	copy(nodes, consenters)
	sort.Slice(nodes, func(i, j int) bool {
		return nodes[i] < nodes[j]
	})
	for i, n := range nodes {
		if i == len(nodes)-1 {
			break
		}
		if n == nodes[i+1] {
			return errors.Errorf("duplicate node %s name in consenter change list", n)
		}
	}
	return nil
}

// RemoveConsenter removes multiple nodes from consensus cluster in batch.
// The removed nodes will create new root zone instance as single node.
func RemoveConsenter(client *client.GatewayClient, chain string, consensusNode string,
	consenter ...Consenter) (*Result, error) {
	if len(chain) == 0 {
		return nil, errors.Wrap(ErrEmptyChainID, "failed to remove consenter")
	}
	if len(consensusNode) == 0 {
		return nil, errors.New("consensus node should not be empty")
	}
	var consenters []*Consenter
	for i := range consenter {
		consenters = append(consenters, &consenter[i])
	}
	sort.Slice(consenters, func(i, j int) bool {
		return consenters[i].Name < consenters[j].Name
	})
	var consNames []string
	for _, c := range consenters {
		consNames = append(consNames, c.Name)
	}
	if _, ok := client.Nodes[consensusNode]; !ok {
		return nil, errors.Errorf("consensus node %s doesn't exist in local network configuration", consensusNode)
	}
	if len(consNames) == 0 {
		return nil, errors.New("empty consenter list which remove from consensus cluster")
	}
	for _, n := range consNames {
		if len(n) == 0 {
			return nil, errors.New("consenter name is empty")
		}
		if _, ok := client.Nodes[n]; !ok {
			return nil, errors.Errorf("consensus node %s doesn't exist in local network configuration", n)
		}
	}
	result, childCtx := newResult(context.Background())
	a := removeProcessor{
		client:     client,
		chain:      chain,
		existNode:  consensusNode,
		consNames:  consNames,
		consenters: consenters,
		res:        result,
	}

	if err := validateName(consNames); err != nil {
		return nil, err
	}
	go a.remove(childCtx)
	return result, nil
}

type removeProcessor struct {
	client     *client.GatewayClient
	chain      string
	existNode  string
	consNames  []string
	consenters []*Consenter
	res        *Result
	quitChain  bool
}

const (
	retryDuration = 5 * time.Second
)

func (p *removeProcessor) remove(parent context.Context) {
	flow := LoadFlow(p.client)

	ok, err := p.check(flow)
	if err != nil {
		p.res.sendNodeError(p.existNode, err)
	}
	if ok {
		configFlow := flow.Operator().Config().SetChain(p.chain).SetTimeout(timeout).SetConsenterList([]string{p.existNode})
		err = configFlow.RemoveConsenter(ConfigVote, p.consNames...)
		if err != nil {
			p.res.sendNodeError(p.existNode, err)
			return
		}
		p.res.sendChangePending()
		err = configFlow.RemoveConsenter(ConfigComplete, p.consNames...)
		if err != nil {
			p.res.sendNodeError(p.existNode, err)
			return
		}
		p.res.sendChangeComplete()
	}

	if p.checkFinish(parent) {
		return
	}

	if p.quitChain {
		p.removeChain(parent, flow)
	}
	p.res.finish()
}

func (p *removeProcessor) checkFinish(parent context.Context) bool {
	group, ctx := utils.WithContext(parent)
	for i := 0; i < len(p.consenters); i++ {
		cons := p.consenters[i]
		group.Go(cons.Name, func() error {
			node, ok := p.client.GetNode(cons.Name)
			if !ok {
				return errors.Errorf("failed to get node %s config from local network config", cons.Name)
			}
			node.RestPort = int(cons.RestPort)
			restc, err := rest.NewRestClient(node)
			if err != nil {
				return errors.WithMessagef(err, "failed to create rest client for node %s", cons.Name)
			}
			for {
				time.Sleep(retryDuration)
				isConsenter, err := rest.IsConsenter(ctx, restc, p.chain)
				if err != nil && errors.Is(err, rest.ErrCheckConsenterTimeout) {
					continue
				}
				if err != nil {
					return errors.WithMessagef(err, "failed to get node %s role", cons.Name)
				}
				if !isConsenter {
					break
				}
			}
			return nil
		})
	}
	if errs := group.Wait(); len(errs) != 0 {
		p.res.sendAbort(errs, "there are some consenters which failed to convert to peer")
		return true
	}
	return false
}

func (p *removeProcessor) removeChain(parent context.Context, flow *Flow) {
	removeGroup, _ := utils.WithContext(parent)
	for _, c := range p.consNames {
		cons := c
		removeGroup.Go(cons, func() error {
			node := flow.Admin().Chain().SetChain(p.chain).SetTimeout(timeout)
			if err := node.Exist(cons); err != nil {
				return errors.WithMessagef(err, "there is not the chain %s in node %s", p.chain, cons)
			}
			if err := node.SetNode(cons).DeleteChain(); err != nil {
				return errors.WithMessagef(err, "failed to remove chain for %s with chain id %s", cons, p.chain)
			}
			return nil
		})
	}
	if errs := removeGroup.Wait(); len(errs) != 0 {
		p.res.sendAbort(errs, "there are some consensus nodes which failed to remove chain")
	}
	p.res.sendRemove()
}

type addProcessor struct {
	client     *client.GatewayClient
	chain      string
	existNode  string
	seeds      []string
	consenters []Consenter
	res        *Result
	joinChain  bool
}

var stepNum uint64 = 10
var timeout int64 = 60

func (p *addProcessor) add(parent context.Context) {
	flow := LoadFlow(p.client)

	ok, err := p.check(flow)
	if err != nil {
		p.res.sendNodeError(p.existNode, err)
		return
	}
	if ok {
		target, ok := p.join(parent, flow)
		if !ok {
			return
		}

		if ok := p.sync(parent, flow, target); !ok {
			return
		}

		configFlow := flow.Operator().Config().SetChain(p.chain).SetTimeout(timeout).SetConsenterList([]string{p.existNode})
		err = configFlow.AddConsenter(ConfigVote, p.consenters...)
		if err != nil {
			p.res.sendNodeError(p.existNode, err)
			return
		}
		p.res.sendChangePending()
		err = configFlow.AddConsenter(ConfigComplete, p.consenters...)
		if err != nil {
			p.res.sendNodeError(p.existNode, err)
			return
		}
		p.res.sendChangeComplete()
	}

	if p.checkFinish(parent) {
		return
	}

	p.res.finish()
}

func (p *addProcessor) checkFinish(parent context.Context) bool {
	group, ctx := utils.WithContext(parent)
	for i := 0; i < len(p.consenters); i++ {
		cons := &p.consenters[i]
		group.Go(cons.Name, func() error {
			node, ok := p.client.GetNode(cons.Name)
			if !ok {
				return errors.Errorf("failed to get node %s config from local network config", cons.Name)
			}
			node.RestPort = int(cons.RestPort)
			restc, err := rest.NewRestClient(node)
			if err != nil {
				return errors.WithMessagef(err, "failed to create rest client for node %s", cons.Name)
			}
			for {
				time.Sleep(retryDuration)
				isConsenter, err := rest.IsConsenter(ctx, restc, p.chain)
				if err != nil && errors.Is(err, rest.ErrCheckConsenterTimeout) {
					continue
				}
				if err != nil {
					return errors.WithMessagef(err, "failed to get node %s role", cons.Name)
				}
				if isConsenter {
					break
				}
			}
			return nil
		})
	}
	if errs := group.Wait(); len(errs) != 0 {
		p.res.sendAbort(errs, "there are some peers which failed to convert to consenter")
		return true
	}
	return false
}

func (p *addProcessor) check(flow *Flow) (bool, error) {
	chainConf, err := flow.Admin().Chain().SetChain(p.chain).SetTimeout(timeout).SetNode(p.existNode).ChainConf()
	if err != nil {
		return false, errors.WithMessagef(err, "failed to fetch latest chain config from exist node %s", p.existNode)
	}
	var cons schema.SysCons
	if err := proto.Unmarshal(chainConf.Config.Consensus, &cons); err != nil {
		return false, errors.WithMessagef(err, "failed to parse consensus configuration "+
			"which extracted from chain configuration fetched from node %s", p.existNode)
	}
	consMap := make(map[string]struct{}, len(cons.Consenter))
	for _, c := range cons.Consenter {
		consMap[c.Name] = struct{}{}
	}

	for i := range p.consenters {
		if _, ok := consMap[p.consenters[i].Name]; ok {
			return false, nil
		}
	}
	return true, nil
}

func (p *removeProcessor) check(flow *Flow) (bool, error) {
	chainConf, err := flow.Admin().Chain().SetChain(p.chain).SetTimeout(timeout).SetNode(p.existNode).ChainConf()
	if err != nil {
		return false, errors.WithMessagef(err, "failed to fetch latest chain config from exist node %s", p.existNode)
	}
	var cons schema.SysCons
	if err := proto.Unmarshal(chainConf.Config.Consensus, &cons); err != nil {
		return false, errors.WithMessagef(err, "failed to parse consensus configuration "+
			"which extracted from chain configuration fetched from node %s", p.existNode)
	}
	consMap := make(map[string]struct{}, len(cons.Consenter))
	for _, c := range cons.Consenter {
		consMap[c.Name] = struct{}{}
	}

	for _, c := range p.consenters {
		if _, ok := consMap[c.Name]; !ok {
			return false, nil
		}
	}
	return true, nil
}

func (p *addProcessor) join(parent context.Context, flow *Flow) (uint64, bool) {
	joinGroup, _ := utils.WithContext(parent)
	for i := 0; i < len(p.consenters); i++ {
		cons := &p.consenters[i]
		joinGroup.Go(cons.Name, func() error {
			err := flow.Admin().Chain().SetChain(p.chain).SetTimeout(timeout).Exist(cons.Name)
			if err == nil {
				return nil
			}
			if !p.joinChain {
				return errors.WithMessagef(err, "node %s should join chain %s first", cons.Name, p.chain)
			}

			if err := flow.Admin().Chain().SetChain(p.chain).
				SetEntrypoint("", false, p.seeds).SetTimeout(timeout).SetNode(cons.Name).
				JoinChain(p.existNode); err != nil {
				return errors.WithMessagef(err, "failed to join chain for %s with chain id = %s", cons.Name, p.chain)
			}
			return nil
		})
	}
	if errs := joinGroup.Wait(); len(errs) != 0 {
		p.res.sendAbort(errs, "there are some pending consensus nodes which failed to join chain")
		return 0, false
	}

	target, err := flow.Observer().SetChain(p.chain).SetNode(p.existNode).SetTimeout(timeout).Height()
	if err != nil {
		p.res.sendNodeError(p.existNode, err)
		return 0, false
	}
	p.res.sendJoin(target)
	return target, true
}

func (p *addProcessor) sync(parent context.Context, flow *Flow, target uint64) bool {
	group, ctx := utils.WithContext(parent)
	for i := 0; i < len(p.consenters); i++ {
		cons := &p.consenters[i]
		group.Go(cons.Name, func() error {
			return p.listenBlockHeight(ctx, flow, cons.Name, target)
		})
	}
	if errs := group.Wait(); len(errs) != 0 {
		p.res.sendAbort(errs, "there are some pending consensus nodes which failed to synchronized")
		return false
	}
	p.res.sendSynchronized()
	return true
}

func (p *addProcessor) listenBlockHeight(ctx context.Context, flow *Flow, cons string, target uint64) error {
	ch, err := flow.Observer().SetChain(p.chain).SetNode(cons).SetTimeout(timeout).BlockEvent(ctx)
	if err != nil {
		return err
	}
	for {
		select {
		case res, ok := <-ch:
			if !ok {
				return ctx.Err()
			}
			h, err := res.Get()
			if err != nil {
				return err
			}
			h++
			if h >= target {
				p.res.sendSyncProgress(cons, h, target)
				return nil
			} else if h%stepNum == 0 {
				p.res.sendSyncProgress(cons, h, target)
			}
		case <-ctx.Done():
			return ctx.Err()
		}
	}
}

// Step represents consensus configuration change operation steps.
type Step int32

const (
	// JoinSuccess represents node has joined the chain.
	JoinSuccess Step = iota
	// BlockSyncProgress represents progress of block synchronizing.
	BlockSyncProgress
	// Synchronized represents block has been synchronized.
	Synchronized
	// Voting represents configuration change in voting state.
	Voting // single policy
	// Pending represents configuration change in pending state.
	Pending
	// Completed represents configuration change in completed state.
	Completed
	// Removed represents node has been removed from chain.
	Removed
	// Success represents the operation is executed successfully.
	Success
	// Abort represents the operation has been aborted with some error.
	Abort
)

var stepName = map[int32]string{
	0: "JoinSuccess",
	1: "BlockSyncProgress",
	2: "Synchronized",
	3: "Voting",
	4: "Pending",
	5: "Completed",
	6: "Removed",
	7: "Success",
	8: "Abort",
}

func (s Step) String() string {
	v, ok := stepName[int32(s)]
	if ok {
		return v
	}
	return strconv.Itoa(int(s))
}

// Progress represents progress result of async multiple operations.
type Progress struct {
	Step        Step               `json:"step,omitempty"`
	BlockGap    uint64             `json:"block_gap,omitempty"`
	SyncPercent map[string]float32 `json:"sync_percent,omitempty"`
	Message     string             `json:"message,omitempty"`
	NodeErrs    map[string]error   `json:"node_errs,omitempty"`
}

// Result represents result and context for async operation request.
type Result struct {
	mu     sync.Mutex
	res    chan *Progress
	done   bool
	err    error
	cancel context.CancelFunc
}

func newResult(parent context.Context) (*Result, context.Context) {
	ctx, cancelFunc := context.WithCancel(parent)
	return &Result{
		res:    make(chan *Progress, 1),
		cancel: cancelFunc,
	}, ctx
}

// Cancel operation context.
func (r *Result) Cancel() {
	r.cancel()
}

// Response returns channel for progress result.
func (r *Result) Response() <-chan *Progress {
	return r.res
}

// Err returns abort error.
func (r *Result) Err() error {
	r.mu.Lock()
	err := r.err
	r.mu.Unlock()
	return err
}

func (r *Result) sendJoin(height uint64) {
	p := Progress{
		Step:     JoinSuccess,
		BlockGap: height,
		Message:  "all pending consenters joined into chain",
	}
	r.send(&p)
}

func (r *Result) sendSyncProgress(node string, height uint64, target uint64) {
	var p Progress
	if height >= target {
		p = Progress{
			Step:     BlockSyncProgress,
			BlockGap: target,
			SyncPercent: map[string]float32{
				node: 1,
			},
			Message: fmt.Sprintf("%s's block height is synchronized", node),
		}
	} else {
		pre := float32(height) / float32(target)
		p = Progress{
			Step:     BlockSyncProgress,
			BlockGap: target,
			SyncPercent: map[string]float32{
				node: pre,
			},
			Message: fmt.Sprintf("%s's block height has been synchronized %.2f%%", node, pre),
		}
	}
	r.send(&p)
}

func (r *Result) sendSynchronized() {
	p := Progress{
		Step:    Synchronized,
		Message: "all consenters have been synchronized",
	}
	r.send(&p)
}

func (r *Result) sendChangeComplete() {
	p := Progress{
		Step:    Completed,
		Message: "consensus configuration change has been completed, vote state converts to completed",
	}
	r.send(&p)
}

func (r *Result) sendChangePending() {
	p := Progress{
		Step:    Pending,
		Message: "consensus configuration change has been voted, vote state converts to pending",
	}
	r.send(&p)
}

func (r *Result) sendRemove() {
	p := Progress{
		Step:    Removed,
		Message: "all consenters haven been removed from chain",
	}
	r.send(&p)
}

func (r *Result) send(progress *Progress) {
	r.mu.Lock()
	defer r.mu.Unlock()

	if r.done || r.err != nil {
		return
	}

	defer func() {
		if err := recover(); err != nil {
			consLg.Errorf("Error Progress = %+v, %+v", progress, err)
		}
	}()

	select {
	case r.res <- progress:
	case <-r.res:
		select {
		case r.res <- progress:
		default:
			consLg.Debugf("replace previous progress from result channel with new %+v", progress)
		}
	}
	if progress.Step == Success {
		r.done = true
		close(r.res)
	}
	if progress.Step == Abort {
		r.err = errors.New(progress.Message)
		close(r.res)
	}
}

func (r *Result) sendNodeError(node string, err error) {
	r.sendAbort(map[string]error{node: err}, err.Error())
}

func (r *Result) sendAbort(errs map[string]error, msg string) {
	p := Progress{
		Step:     Abort,
		NodeErrs: errs,
		Message:  msg,
	}
	r.send(&p)
}

func (r *Result) finish() {
	p := Progress{Step: Success}
	r.send(&p)
}
