/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package flow

import (
	"encoding/hex"
	"time"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/sdk/node"
	"huawei.com/huaweichain/sdk/rawmessage"
)

type consenterFlow struct {
	*performerFlow
	consenterList []string
}

// SetConsenterList is used to set consenter node list
func (c *consenterFlow) SetConsenterList(consenterlist []string) (*consenterFlow, error) {
	for _, n := range consenterlist {
		_, ok := c.Flow.stub.gatewayClient.Nodes[n]
		if !ok {
			return nil, errors.Wrapf(ErrNoSuchNode, "consensus node name = %s", n)
		}
	}
	c.consenterList = consenterlist
	return c, nil
}

// Consenter is used to get a usable consenter node
func (c *consenterFlow) Consenter() (*node.WNode, error) {
	n, err := c.randomNode(c.consenterList)
	if err != nil {
		return nil, errors.Wrap(ErrNoAvailableConsenter, err.Error())
	}
	return c.stub.gatewayClient.Nodes[n], nil
}

// SendTx is used to send tx
func (c *consenterFlow) SendTx(tx *rawmessage.TxRawMsg) error {
	consenter, err := c.Consenter()
	if err != nil {
		return err
	}
	txID := hex.EncodeToString(tx.Hash)
	listener, err := consenter.EventAction.GetTxEventService(c.chain)
	if err != nil {
		return errors.WithMessagef(err, "failed to start event listener for chain %s", c.chain)
	}
	defer listener.Close()
	event, err := listener.RegisterTx(tx.Hash)
	if err != nil {
		return errors.WithMessagef(err, "failed to register transaction event which tx id is %s", txID)
	}
	rawResp, err := consenter.ContractAction.Transaction(tx.Msg)
	if err != nil {
		return errors.WithMessagef(err, "failed to send transaction to consensus service which tx id is %s", txID)
	}
	var resp common.Response
	if err := proto.Unmarshal(rawResp.Payload, &resp); err != nil {
		return errors.WithMessagef(err, "failed to unmarshal response from consensus service which tx id is %s", txID)
	}
	if resp.Status != common.SUCCESS {
		return errors.Errorf("consensus response status [%s] with message: %s", resp.Status, resp.StatusInfo)
	}
	select {
	case e := <-event:
		if e == nil {
			return errors.New("transaction event is empty")
		}
		if e.Status != common.VALID {
			return errors.Errorf("failed to validate transaction caused by %s which tx id is %s",
				e.Status, txID)
		}
	case <-time.After(time.Duration(c.timeout) * time.Second):
		return errors.Errorf("transaction event with timeout %ds", c.timeout)
	}
	return nil
}
