/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package flow

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto/common"
)

// ContractFlow is used to handle contract related process
// which includes query and invoke
type ContractFlow struct {
	*endorserFlow
	domains []string
	name    string
}

func newContractFlow(stub *stub) *ContractFlow {
	contractFlowInstance := ContractFlow{
		endorserFlow: &endorserFlow{
			consenterFlow: &consenterFlow{
				performerFlow: &performerFlow{
					ObserverFlow: &ObserverFlow{
						Flow: &Flow{
							stub: stub,
						},
					},
				},
			},
			Org: make(map[string]*orgNodeList),
		},
	}
	return &contractFlowInstance
}

// SetDomains is used to set domain
func (c *ContractFlow) SetDomains(domains []string) *ContractFlow {
	c.domains = domains
	return c
}

// SetChain is used to set chain id
func (c *ContractFlow) SetChain(chain string) *ContractFlow {
	c.chain = chain
	return c
}

// SetNode is used to set node
func (c *ContractFlow) SetNode(node string) *ContractFlow {
	_, err := c.ObserverFlow.SetNodeWithErr(node)
	if err != nil {
		log.Fatal(err.Error())
	}
	return c
}

// SetContractName is used to set contract Name
func (c *ContractFlow) SetContractName(contractName string) *ContractFlow {
	c.name = contractName
	return c
}

// SetConsenterList is used to set consenter list
func (c *ContractFlow) SetConsenterList(consenterList []string) *ContractFlow {
	_, err := c.consenterFlow.SetConsenterList(consenterList)
	if err != nil {
		log.Fatal(err.Error())
	}
	return c
}

// SetTimeout for request timeout.
func (c *ContractFlow) SetTimeout(timeout int64) *ContractFlow {
	c.timeout = timeout
	return c
}

// Invoke is used to send invoke message
func (c *ContractFlow) Invoke(method string, args []string) ([]byte, error) {
	// send invoke message
	helper := c.stub.GetContractRawMessage()
	invocation := helper.BuildContractInvocation(c.name, method, args)
	req, err := helper.BuildInvokeRawMsg(c.chain, invocation, c.domains)
	if err != nil {
		return nil, errors.WithMessage(err, "BuildInvokeRawMsg error")
	}
	var approvals []*common.RawMessage
	res, err := c.Node.ContractAction.Invoke(req)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to endorse for transaction")
	}
	approvals = append(approvals, res)

	// send transaction message
	tx, err := helper.BuildTxRawMsg(approvals)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to construct transaction")
	}
	val, err := helper.ParseQueryResult(res)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to parse endorsement return value")
	}

	if err := c.SendTx(tx); err != nil {
		return nil, errors.WithMessage(err, "failed to send transaction to consensus")
	}
	return val, nil
}

// InvokeWithResponse is used to send endorsement and transaction without return value.
func (c *ContractFlow) InvokeWithResponse(method string, args []string) error {
	return c.invoke(method, args)
}

func (c *ContractFlow) invoke(method string, args []string) error {
	// send invoke message
	cv := c.stub.gatewayClient.ContractRawMessage.BuildContractInvocation(c.name, method, args)
	invokeMsg, err := c.stub.GetContractRawMessage().BuildInvokeRawMsg(c.chain, cv, c.domains)
	if err != nil {
		return errors.WithMessage(err, "BuildInvokeRawMsg error")
	}
	var invokeResponses []*common.RawMessage
	// org init
	err = c.orgInit()
	if err != nil {
		return errors.WithMessage(err, "orgInit error")
	}
	// select one node for each org
	for orgName := range c.Org {
		response, errNew := c.orgInvoke(orgName, invokeMsg)
		if errNew != nil {
			return errors.WithMessage(errNew, "org check fail")
		}
		invokeResponses = append(invokeResponses, response)
	}
	// send transaction message
	rawTxMessage, err := c.stub.GetContractRawMessage().BuildTxRawMsg(invokeResponses)
	if err != nil {
		return errors.WithMessage(err, "BuildTxRawMsg error")
	}
	err = c.SendTx(rawTxMessage)
	if err != nil {
		return errors.WithMessage(err, "SendTx error")
	}
	return nil
}

func (c *ContractFlow) orgInvoke(orgName string, message *common.RawMessage) (*common.RawMessage, error) {
	endorserNode, err := c.GetOrgEndorser(orgName)
	if err != nil {
		return nil, errors.New("nodes unconnected")
	}
	response, err := c.stub.gatewayClient.Nodes[endorserNode].ContractAction.Invoke(message)
	if err != nil {
		return nil, errors.WithMessage(err, "Invoke error")
	}
	return response, nil
}

// Query is used to query invoke message
func (c *ContractFlow) Query(method string, args []string) ([]byte, error) {
	cv := c.stub.GetContractRawMessage().BuildContractInvocation(c.name, method, args)
	invokeMsg, err := c.stub.GetContractRawMessage().BuildInvokeRawMsg(c.chain, cv, c.domains)
	if err != nil {
		return nil, errors.WithMessage(err, "BuildInvokeRawMsg error")
	}
	if c.Node == nil {
		return nil, errors.WithMessage(err, "need to set node")
	}
	invokeResponse, err := c.Node.ContractAction.Query(invokeMsg)
	if err != nil {
		return nil, errors.WithMessage(err, "error in contract action query")
	}
	result, err := c.stub.GetContractRawMessage().ParseQueryResult(invokeResponse)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to parse the result from query response")
	}
	return result, nil
}

// ContractType is used to pre-define the contract
type ContractType struct {
	sandbox  string
	language string
}

// GetContractType is used to get contract type
func GetContractType(contractType string) *ContractType {
	switch contractType {
	case "DockerJava":
		return generateContractType("docker", "java")
	case "Native":
		return generateContractType("native", "")
	case "Tee":
		return generateContractType("teewasm", "bytecode")
	case "Wasm":
		return generateContractType("nativewasm", "bytecode")
	case "Fabric":
		return generateContractType("docker", "bytecode")
	case "Evm":
		return generateContractType("docker", "go")
	default:
		return generateContractType("docker", "go")
	}
}

// GetSandbox is used to get sandbox
func (c *ContractType) GetSandbox() string {
	return c.sandbox
}

// GetLanguage is used to get language
func (c *ContractType) GetLanguage() string {
	return c.language
}

func isFabric(ct *ContractType) bool {
	if ct.sandbox == "docker" && ct.language == "go" {
		return true
	}
	return false
}

func generateContractType(sandbox, language string) *ContractType {
	return &ContractType{
		sandbox:  sandbox,
		language: language,
	}
}
