/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

// Package contract provides client contract caller.
package contract

import (
	"sync"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/sdk/client"
	"huawei.com/huaweichain/sdk/flow"
)

type caller struct {
	mu        sync.RWMutex
	flow      *flow.Flow
	chain     string
	endorsers func() string
	consenter func() string
	listener  func() string
}

// Flow returns inner flow instance.
func (c *caller) Flow() *flow.Flow {
	return c.flow
}

// ChainID returns chain id for contract.
func (c *caller) ChainID() string {
	c.mu.RLock()
	chain := c.chain
	c.mu.RUnlock()
	return chain
}

// Endorser returns endorser for transaction.
func (c *caller) Endorser() string {
	c.mu.RLock()
	e := c.endorsers()
	c.mu.RUnlock()
	return e
}

// Consenter returns consenter for transaction.
func (c *caller) Consenter() string {
	c.mu.RLock()
	cons := c.consenter()
	c.mu.RUnlock()
	return cons
}

// Listener returns node of listener for transaction result.
func (c *caller) Listener() string {
	c.mu.RLock()
	l := c.listener()
	c.mu.RUnlock()
	return l
}

// DefaultContractCaller creates a Contract Caller for client sdk calling contract.
func DefaultContractCaller(client *client.GatewayClient, chain string,
	endorser, consenter []string) (Caller, error) {
	if client == nil {
		return nil, errors.New("client is nil")
	}
	if len(chain) == 0 {
		return nil, errors.New("chain id is empty")
	}
	if len(endorser) == 0 || len(consenter) == 0 {
		return nil, errors.New("empty endorser or consenter")
	}
	return &caller{
		flow:  flow.LoadFlow(client),
		chain: chain,
		endorsers: func() string {
			return endorser[0]
		},
		consenter: func() string {
			return consenter[0]
		},
		listener: func() string {
			return endorser[0]
		},
	}, nil
}

// Caller represents a caller of contract.
type Caller interface {
	Flow() *flow.Flow
	ChainID() string
	Endorser() string
	Consenter() string
	Listener() string
}
