/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

// Package evm provides function wrappers of evm contract function for client caller.
package evm

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/sdk/flow/contract"
)

const (
	// NativeContractEVM represents contract name of evm.
	NativeContractEVM = "NATIVE_CUSTOM_EVM"

	methodAccount  = "account"
	methodDeploy   = "deploy"
	methodBytecode = "bytecode"
	methodCall     = "call"
)

const (
	timeout = 60
)

// Transact represents contract function name for executing encoded command for evm transaction.
func Transact(caller contract.Caller, address, encodedInput string) ([]byte, error) {
	val, err := caller.Flow().User().Contract().SetChain(caller.ChainID()).SetContractName(NativeContractEVM).
		SetNode(caller.Endorser()).SetConsenterList([]string{caller.Consenter()}).SetTimeout(timeout).
		Invoke(methodCall, []string{address, encodedInput})
	if err != nil {
		return nil, errors.WithMessage(err, "failed to transact evm contract with encoded input")
	}
	return val, nil
}

// Call represents contract function name for executing encoded command for evm.
func Call(caller contract.Caller, address, encodedInput string) ([]byte, error) {
	res, err := caller.Flow().User().Contract().SetChain(caller.ChainID()).SetContractName(NativeContractEVM).
		SetNode(caller.Endorser()).SetTimeout(timeout).
		Query(methodCall, []string{address, encodedInput})
	if err != nil {
		return nil, errors.WithMessage(err, "failed to call evm contract with encoded input")
	}
	return res, nil
}

// Deploy represents contract function name for deploying compiled contract bytecode.
func Deploy(caller contract.Caller, bytecode string) (string, error) {
	addr, err := caller.Flow().User().Contract().SetChain(caller.ChainID()).SetContractName(NativeContractEVM).
		SetNode(caller.Endorser()).SetConsenterList([]string{caller.Consenter()}).SetTimeout(timeout).
		Invoke(methodDeploy, []string{bytecode})
	if err != nil {
		return "", errors.WithMessage(err, "failed to deploy contract bytecode to evm")
	}
	return string(addr), nil
}

// Bytecode represents contract function name for querying deployed executable bytecode of specific contract address.
func Bytecode(caller contract.Caller, address string) ([]byte, error) {
	res, err := caller.Flow().User().Contract().SetChain(caller.ChainID()).SetContractName(NativeContractEVM).
		SetNode(caller.Endorser()).SetTimeout(timeout).
		Query(methodBytecode, []string{address})
	if err != nil {
		return nil, errors.WithMessage(err, "failed to fetch contract bytecode from evm")
	}
	return res, nil
}

// Account represents contract function name for querying account address of caller.
func Account(caller contract.Caller) ([]byte, error) {
	res, err := caller.Flow().User().Contract().SetChain(caller.ChainID()).SetContractName(NativeContractEVM).
		SetNode(caller.Endorser()).SetTimeout(timeout).
		Query(methodAccount, nil)
	if err != nil {
		return nil, errors.WithMessage(err, "failed to fetch account address from evm")
	}
	return res, nil
}
