/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

// Package flow is the implementation of flow sdk for go client.
package flow

import (
	"errors"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/sdk/client"
	"huawei.com/huaweichain/sdk/utils"
)

var (
	// ErrNoSuchNode returns when no such node exists in local network configuration.
	ErrNoSuchNode = errors.New("no such node exists in local network configuration")
	// ErrNoAvailableConsenter returns when no available consenter for transaction consensus step.
	ErrNoAvailableConsenter = errors.New("no available consenter for transaction consensus step")
	// ErrNoAvailableEndorser returns when no available endorser for transaction endorsing step.
	ErrNoAvailableEndorser = errors.New("no available endorser for transaction endorsing step")

	// ErrEmptyChainID returns when chain id is empty.
	ErrEmptyChainID = errors.New("chain id is empty")
)

// Flow is the root struct
type Flow struct {
	stub    *stub
	timeout int64
}

var log = logger.GetModuleLogger("sdk.flow", "default")

const oneMin int64 = 60

// MakeEmpFlow is used to create a empty flow instance
func MakeEmpFlow() *Flow {
	return &Flow{
		stub:    newStub(),
		timeout: oneMin,
	}
}

// LoadFlow creates a flow instance from manual client sdk.
func LoadFlow(client *client.GatewayClient) *Flow {
	return &Flow{
		stub: &stub{
			gatewayClient: client,
		},
		timeout: int64(utils.GetTimeout()),
	}
}

// SetTimeout is used to set sdk waiting timeout
func (f *Flow) SetTimeout(timeout int64) *Flow {
	f.timeout = timeout
	return f
}

// Identity is used to return a IdentityFlow instance
func (f *Flow) Identity() *IdentityFlow {
	return newIdentityFlow(f.stub)
}

// TLS is used to return a TlSFlow instance
func (f *Flow) TLS() *TlSFlow {
	return newTLSFlow(f.stub)
}

// Net is used to return a NetFlow instance
func (f *Flow) Net() *NetFlow {
	return newNetflow(f.stub)
}

// Observer is used to return a ObserverFlow instance
func (f *Flow) Observer() *ObserverFlow {
	return newObserverFlow(f.stub)
}

// Admin is used to return a AdminFlow instance
func (f *Flow) Admin() *AdminFlow {
	return newAdminFlow(f.stub)
}

// Operator is used to return a OperatorFlow instance
func (f *Flow) Operator() *OperatorFlow {
	return newOperatorFlow(f.stub)
}

// User is used to return a UserFlow instance
func (f *Flow) User() *UserFlow {
	return newUserFlow(f.stub)
}
