/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package flow

import (
	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/sdk/rawmessage"
)

// LifecycleFlow is used to handle the contract lifecycle operation
// which includes import, vote, freeze
type LifecycleFlow struct {
	*consenterFlow
	contractName   string
	packagePath    string
	version        string
	contract       *rawmessage.Contract
	historySupport bool
	initRequired   bool
	contractType   *ContractType
}

func newLifecycleFlow(stub *stub) *LifecycleFlow {
	lifecycleFlowInstance := LifecycleFlow{
		consenterFlow: &consenterFlow{
			performerFlow: &performerFlow{
				ObserverFlow: &ObserverFlow{
					Flow: &Flow{
						stub: stub,
					},
				},
			},
		},
	}
	return &lifecycleFlowInstance
}

// Define is used to set parameters for LifecycleFlow
func (l *LifecycleFlow) Define(version string, pkgPath string) *LifecycleFlow {
	l.version = version
	l.packagePath = pkgPath
	return l
}

// SetContractName is used to set contractName for LifecycleFlow
func (l *LifecycleFlow) SetContractName(name string) *LifecycleFlow {
	l.contractName = name
	return l
}

// SetTimeout is used to set timeout for LifecycleFlow
func (l *LifecycleFlow) SetTimeout(timeout int64) *LifecycleFlow {
	l.timeout = timeout
	return l
}

// SetHistory is used to enable historySupport or not
func (l *LifecycleFlow) SetHistory(history bool) *LifecycleFlow {
	l.historySupport = history
	return l
}

// SetInit is used to enable initRequired
func (l *LifecycleFlow) SetInit(init bool) *LifecycleFlow {
	l.initRequired = init
	return l
}

// SetContract is used to set contract
func (l *LifecycleFlow) SetContract(chainID string, contractName string, version string) *LifecycleFlow {
	l.contract = rawmessage.NewContract(chainID, contractName, version)
	return l
}

// SetContractType is used to set contractType
func (l *LifecycleFlow) SetContractType(contractType string) *LifecycleFlow {
	l.contractType = GetContractType(contractType)
	return l
}

// SetPkgPath is used to set pkgPath
func (l *LifecycleFlow) SetPkgPath(pkgPath string) *LifecycleFlow {
	l.packagePath = pkgPath
	return l
}

// Install is used to import contract
func (l *LifecycleFlow) Install(node string) error {
	l.stub.gatewayClient.SetTimeout(l.timeout)
	var rawMessage *common.RawMessage
	var err error
	if l.contractType == nil {
		return errors.New("error in install - need to set contract type previously")
	}
	if isFabric(l.contractType) {
		rawMessage, err = l.stub.GetLifecycleRawMessage().BuildImportFabricRawMessage(l.contract,
			l.packagePath, l.contractType.GetSandbox(), l.contractType.GetLanguage())
	} else {
		rawMessage, err = l.stub.GetLifecycleRawMessage().BuildImportRawMessage(l.contract,
			l.packagePath, l.contractType.GetSandbox(), l.contractType.GetLanguage())
	}
	if err != nil {
		return errors.WithMessage(err, "build lifecycle import raw message error")
	}
	// send raw message
	nodeMap := l.stub.gatewayClient.Nodes
	nodeAction := nodeMap[node]
	responseMsg, err := nodeAction.ContractAction.ContractImport(rawMessage)
	if err != nil {
		return errors.WithMessage(err, "contract action import error")
	}
	response := &common.Response{}
	err = proto.Unmarshal(responseMsg.Payload, response)
	if err != nil {
		return errors.WithMessage(err, "errors occur in proto unmarshal")
	}
	PrintResponse(response, node, response.StatusInfo)
	if response.Status != common.SUCCESS {
		return errors.Errorf("isntall response status error, Status: %v, Statusinfo: %v",
			response.Status.String(), response.StatusInfo)
	}
	return nil
}

// Uninstall is used to  unImport contract
func (l *LifecycleFlow) Uninstall() error {
	l.stub.gatewayClient.SetTimeout(l.timeout)
	rawMessage, err := l.stub.GetLifecycleRawMessage().BuildUnImportRawMessage(l.contract)
	if err != nil {
		return errors.WithMessage(err, "errors occur in building unimport raw message")
	}
	if l.Node == nil {
		return errors.New("need to set node before")
	}
	responseMsg, err := l.Node.ContractAction.ContractUnImport(rawMessage)
	if err != nil {
		return errors.WithMessage(err, "contract action unimport error")
	}
	response := &common.Response{}
	err = proto.Unmarshal(responseMsg.Payload, response)
	if err != nil {
		return errors.WithMessage(err, "errors occur in proto unmarshal")
	}
	PrintResponse(response, l.Node.ID, response.StatusInfo)
	if response.Status != common.SUCCESS {
		return errors.Errorf("uninstall response status error, Status: %v, Statusinfo: %v",
			response.Status.String(), response.StatusInfo)
	}
	return nil
}

// Freeze is used to freeze contract
func (l *LifecycleFlow) Freeze(contract string) error {
	return l.manage(contract, "freeze")
}

// Unfreeze is used to unfreeze contract
func (l *LifecycleFlow) Unfreeze(contract string) error {
	return l.manage(contract, "unfreeze")
}

// Remove is used to remove contract
func (l *LifecycleFlow) Remove(contract string) error {
	return l.manage(contract, "destroy")
}

func (l *LifecycleFlow) manage(contract string, option string) error {
	rawMessage, err := l.stub.GetLifecycleRawMessage().BuildManageRawMessage(l.chain, contract, option)
	if err != nil {
		return errors.WithMessage(err, "errors occur in building manage raw message")
	}
	err = l.SendTx(rawMessage)
	if err != nil {
		return errors.WithMessage(err, "errors occur in send tx")
	}
	return nil
}

// Vote is used to vote
func (l *LifecycleFlow) Vote(description string, policy string) error {
	rawMessage, err := l.stub.GetLifecycleRawMessage().BuildVoteRawMessage(l.contract,
		description, policy, l.historySupport, l.initRequired)
	if err != nil {
		return errors.WithMessage(err, "errors occur in building vote raw message")
	}
	err = l.SendTx(rawMessage)
	if err != nil {
		return errors.WithMessage(err, "errors occur in sendTx")
	}
	return nil
}

// SetConsenterList is used to set consenterList
func (l *LifecycleFlow) SetConsenterList(consenter []string) *LifecycleFlow {
	_, err := l.consenterFlow.SetConsenterList(consenter)
	if err != nil {
		log.Fatal(err.Error())
	}
	return l
}

// SetChain is used to set chainId
func (l *LifecycleFlow) SetChain(chain string) *LifecycleFlow {
	l.chain = chain
	return l
}

// SetNode is used to set node
func (l *LifecycleFlow) SetNode(node string) *LifecycleFlow {
	_, err := l.ObserverFlow.SetNodeWithErr(node)
	if err != nil {
		log.Fatal(err.Error())
	}
	return l
}

// QueryContractState is used to query contract state
func (l *LifecycleFlow) QueryContractState(contract string) error {
	var err error
	l.stub.gatewayClient.SetTimeout(l.timeout)
	rawMessage, err := l.stub.GetLifecycleRawMessage().BuildQueryStateRawMessage(l.chain, contract)
	if err != nil {
		return errors.WithMessage(err, "errors occur in building query contract state raw message")
	}
	if l.Node == nil {
		return errors.New("need to set node before")
	}
	responseMsg, err := l.Node.ContractAction.QueryState(rawMessage)
	if err != nil {
		return errors.WithMessage(err, "errors occur in query state action ")
	}
	response := &common.Response{}
	if err := proto.Unmarshal(responseMsg.Payload, response); err != nil {
		return errors.WithMessage(err, "unmarshal response error")
	}
	PrintResponse(response, l.Node.ID, response.StatusInfo)
	if response.Status != common.SUCCESS {
		return errors.Errorf("response error: status: %v, info: %v", response.Status.String(), response.StatusInfo)
	}
	return nil
}
