/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package flow

import (
	"context"
	"encoding/hex"
	"encoding/json"
	"strconv"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/action/event"
	"huawei.com/huaweichain/sdk/node"
)

// ObserverFlow is the role of observer
type ObserverFlow struct {
	*Flow
	chain string
	Node  *node.WNode
}

func newObserverFlow(stub *stub) *ObserverFlow {
	return &ObserverFlow{
		Flow: &Flow{
			stub: stub,
		},
		chain: "",
		Node:  &node.WNode{},
	}
}

// SetChain is used to set chain
func (o *ObserverFlow) SetChain(chain string) *ObserverFlow {
	o.chain = chain
	return o
}

// CheckNodeExisted is used to check node existed
func (o *ObserverFlow) CheckNodeExisted(node string) bool {
	return o.stub.gatewayClient.Nodes[node] != nil
}

// SetNodeWithErr is used to set node
func (o *ObserverFlow) SetNodeWithErr(node string) (*ObserverFlow, error) {
	n, ok := o.stub.gatewayClient.Nodes[node]
	if !ok {
		return nil, errors.Wrap(ErrNoSuchNode, "failed to set node for observer flow")
	}
	o.Node = n
	return o, nil
}

// SetNode is used to set node without error return
func (o *ObserverFlow) SetNode(node string) *ObserverFlow {
	_, err := o.SetNodeWithErr(node)
	if err != nil {
		log.Panicf("%+v", err)
	}
	return o
}

// SetTimeout for request.
func (o *ObserverFlow) SetTimeout(timeout int64) *ObserverFlow {
	o.timeout = timeout
	return o
}

// QueryGenesis is used to query genesis block
func (o *ObserverFlow) QueryGenesis() error {
	return o.QueryBlock(0)
}

// QueryBlock is used to query block through block num
func (o *ObserverFlow) QueryBlock(num uint64) error {
	o.stub.gatewayClient.SetTimeout(o.timeout)
	rawMsg, err := o.stub.GetQueryRawMessage().BuildBlockRawMessage(o.chain, num)
	if err != nil {
		return errors.WithMessage(err, "build create chain raw message error")
	}
	responseRaw, err := o.Node.QueryAction.GetBlockByNum(rawMsg)
	if err != nil {
		return errors.WithMessage(err, "query action query block by blockheight error")
	}
	response := &common.Response{}
	if err = proto.Unmarshal(responseRaw.Payload, response); err != nil {
		return errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if response.Status != common.SUCCESS {
		PrintResponse(response, o.Node.ID, "")
		return errors.Errorf("response error: status: %v, info: %v", response.Status.String(), response.StatusInfo)
	}
	block := &common.Block{}
	if err = proto.Unmarshal(response.Payload, block); err != nil {
		return errors.WithMessage(err, "unmarshal block error")
	}
	blockMap := make(map[string]string)
	blockNumber := "block number"
	parentHash := "parent hash"
	bodyHash := "body hash"
	blockMap[blockNumber] = strconv.Itoa(int(block.Header.Number))
	blockMap[parentHash] = Hash2str(block.Header.ParentHash)
	blockMap[bodyHash] = Hash2str(block.Header.BodyHash)
	b, err := json.Marshal(blockMap)
	if err != nil {
		return errors.WithMessage(err, "json.Marshal failed")
	}
	PrintResponse(response, o.Node.ID, string(b))
	return nil
}

// NumberResult contains block number or error.
type NumberResult struct {
	height uint64
	err    error
}

// Get current block number.
func (r *NumberResult) Get() (uint64, error) {
	return r.height, r.err
}

// BlockEvent listens block number event and return result channel immediately.
func (o *ObserverFlow) BlockEvent(ctx context.Context) (<-chan NumberResult, error) {
	service, err := o.Node.EventAction.GetBlockEventService(o.chain)
	if err != nil {
		return nil, err
	}
	iter, err := service.RegisterBlockNumEvent()
	if err != nil {
		return nil, err
	}
	ch := make(chan NumberResult)
	go listenBlockEvent(ctx, ch, service, iter)
	return ch, nil
}

func listenBlockEvent(ctx context.Context, ch chan NumberResult, service *event.BlockEventService,
	iter *event.BlockNumIterator) {
	defer func() {
		close(ch)
		service.Close()
	}()
	for {
		next, err := iter.Next() // FIXME: 是否有超时机制保护，避免协程泄漏
		if err != nil {
			select {
			case ch <- NumberResult{err: err}:
			case <-ctx.Done():
			}
			return
		}
		select {
		case ch <- NumberResult{height: next}:
		case <-ctx.Done():
			return
		}
	}
}

// Height is used to query chain state
func (o *ObserverFlow) Height() (uint64, error) {
	o.stub.gatewayClient.SetTimeout(o.timeout)
	req, err := o.stub.GetQueryRawMessage().BuildLatestChainStateRawMessage(o.chain)
	if err != nil {
		return 0, errors.WithMessage(err, "build query chain state raw message error")
	}
	rawRes, err := o.Node.QueryAction.GetLatestChainState(req)
	if err != nil {
		return 0, errors.WithMessage(err, "query action get latest chain state error")
	}
	var res common.Response
	if err = proto.Unmarshal(rawRes.Payload, &res); err != nil {
		return 0, errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if res.Status != common.SUCCESS {
		return 0, errors.Errorf("error_code = %d, failed to fetch latest chain configuration caused by %s",
			res.Status, res.StatusInfo)
	}
	var state nodeservice.LatestChainState
	if err := proto.Unmarshal(res.Payload, &state); err != nil {
		return 0, errors.Wrap(err, "failed to parse chain configuration data")
	}

	return state.Height, nil
}

// QueryTx is used to query tx
func (o *ObserverFlow) QueryTx(txHash []byte) error {
	o.stub.gatewayClient.SetTimeout(o.timeout)
	rawMsg, err := o.stub.GetQueryRawMessage().BuildTxRawMessage(o.chain, txHash)
	if err != nil {
		return errors.WithMessage(err, "build query tx raw message error")
	}
	responseRaw, err := o.Node.QueryAction.GetTxByHash(rawMsg)
	if err != nil {
		return errors.WithMessage(err, "query action gettxbyhash error")
	}
	response := &common.Response{}
	if err = proto.Unmarshal(responseRaw.Payload, response); err != nil {
		return errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if response.Status != common.SUCCESS {
		PrintResponse(response, o.Node.ID, "")
		return errors.Errorf("response error: status: %v, info: %v", response.Status.String(), response.StatusInfo)
	}
	tx := &common.Tx{}
	if err := proto.Unmarshal(response.Payload, tx); err != nil {
		return errors.WithMessage(err, "unmarshal tx error")
	}
	txPayload := &common.TxPayload{}
	if err := proto.Unmarshal(tx.GetFull().Payload, txPayload); err != nil {
		return errors.WithMessage(err, "unmarshal tx payload error")
	}
	PrintResponse(response, o.Node.ID, hex.EncodeToString(tx.Hash))
	return nil
}

// QueryTxResult is used to query tx result
func (o *ObserverFlow) QueryTxResult(txHash []byte) error {
	o.stub.gatewayClient.SetTimeout(o.timeout)
	rawMsg, err := o.stub.GetQueryRawMessage().BuildTxRawMessage(o.chain, txHash)
	if err != nil {
		return errors.WithMessage(err, "build query tx result raw message error")
	}
	responseRaw, err := o.Node.QueryAction.GetTxResultByTxHash(rawMsg)
	if err != nil {
		return errors.WithMessage(err, "query action gettxresultbytxhash error")
	}
	response := &common.Response{}
	if err = proto.Unmarshal(responseRaw.Payload, response); err != nil {
		return errors.WithMessage(err, "unmarshal responseraw error: %v")
	}
	if response.Status != common.SUCCESS {
		PrintResponse(response, o.Node.ID, "")
		return errors.Errorf("response error: status: %v, info: %v", response.Status.String(), response.StatusInfo)
	}
	txResult := &common.TxResult{}
	if err := proto.Unmarshal(response.Payload, txResult); err != nil {
		return errors.WithMessage(err, "unmarshal transaction error")
	}
	PrintResponse(response, o.Node.ID, txResult.Status.String())
	return nil
}
