/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

// Package rest provides restful api client for querying information about chain and node.
package rest

import (
	"context"
	"crypto/tls"
	"fmt"
	"net/http"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/logger"
	"huawei.com/huaweichain/sdk/client"
	"huawei.com/huaweichain/sdk/node"
)

var lg = logger.GetModuleLogger("sdk.rest", "")

// Client wrappers http.Client for rest api client.
type Client struct {
	*http.Client
	protocol string
	host     string
	port     int
}

// NewClient creates a rest client instance by GatewayClient and specific node name.
func NewClient(sdk *client.GatewayClient, nodeName string) (*Client, error) {
	n, ok := sdk.GetNode(nodeName)
	if !ok {
		return nil, errors.Errorf("node %s doesn't exsit in local network configuration", nodeName)
	}

	return NewRestClient(n)
}

// NewRestClient creates a new rest http client for node rest api.
func NewRestClient(node *node.WNode) (*Client, error) {
	if node.RestPort == 0 {
		return nil, errors.Errorf("failed to initialize rest client for node %s "+
			"without rest port in local configuration", node.ID)
	}
	var protocol string
	var tlsConf *tls.Config
	if node.TLSEnable() {
		c, err := node.ExportTLSConfig()
		if err != nil {
			return nil, errors.WithMessagef(err, "failed to export http tls configuration from node %s", node.ID)
		}
		tlsConf = c
		protocol = "https"
	} else {
		println("disable tls")
		protocol = "http"
	}
	return &Client{
		Client: &http.Client{
			Transport: &http.Transport{
				TLSClientConfig:   tlsConf,
				DisableKeepAlives: true,
			},
		},
		protocol: protocol,
		host:     node.HostOverride,
		port:     node.RestPort,
	}, nil
}

// Send http GET request by rest api path.
func (c *Client) Send(ctx context.Context, path string) (*http.Response, error) {
	url := fmt.Sprintf("%s://%s:%d%s", c.protocol, c.host, c.port, path)
	println(url)
	request, err := http.NewRequestWithContext(ctx, http.MethodGet, url, nil)
	if err != nil {
		return nil, err
	}
	request.Header.Set("Content-Type", "application/json")
	return c.Do(request)
}
