/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package rest

import (
	"context"
	"encoding/json"
	goerrors "errors"
	"fmt"
	"io/ioutil"
	"net/http"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/sdk/client"
)

// ErrCheckConsenterTimeout returns timeout error when check node whether consensus node.
var ErrCheckConsenterTimeout = goerrors.New("node is busy, check consenter timeout")

// Node represents rest client for fetching information about node.
type Node struct {
	sdk *client.GatewayClient
}

// NewNode creates an instance of rest client about node info.
func NewNode(sdk *client.GatewayClient) *Node {
	return &Node{sdk: sdk}
}

// Payload represents response json payload data structure.
type Payload struct {
	IsConsenter bool `json:"is_consenter"`
}

// IsConsenter checks whether the node is consenter for specific chain.
func (n *Node) IsConsenter(ctx context.Context, node string, chain string) (bool, error) {
	c, err := NewClient(n.sdk, node)
	if err != nil {
		return false, err
	}
	return IsConsenter(ctx, c, chain)
}

// IsConsenter checks whether the node is consenter for specific chain.
func IsConsenter(ctx context.Context, c *Client, chain string) (bool, error) {
	defer c.CloseIdleConnections()
	path := fmt.Sprintf("/node/stat?chain=%s", chain)
	resp, err := c.Send(ctx, path)
	if err != nil {
		return false, err
	}
	defer func() {
		if cerr := resp.Body.Close(); cerr != nil {
			lg.Error("failed to close request body")
		}
	}()
	if resp.StatusCode == http.StatusInternalServerError {
		msg, _ := ioutil.ReadAll(resp.Body)
		return false, errors.Wrapf(ErrCheckConsenterTimeout, "failed to fetch node role from rest api with message %s", msg)
	}
	if resp.StatusCode != http.StatusOK {
		msg, _ := ioutil.ReadAll(resp.Body)
		return false, errors.Errorf("failed to fetch node role from rest api "+
			"with code %d and message %s", resp.StatusCode, msg)
	}
	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return false, errors.Wrapf(err, "failed to read response body with %s", resp.Body)
	}
	var p Payload
	if err := json.Unmarshal(body, &p); err != nil {
		return false, errors.Wrapf(err, "failed to unmarshal json payload with %s", body)
	}
	return p.IsConsenter, nil
}
