/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package test

import (
	"os"
	"testing"

	"github.com/stretchr/testify/assert"

	"huawei.com/huaweichain/sdk/flow"
)

// mock server for testing chain operation
func TestChainOperation(t *testing.T) {
	// open mock server
	freePort, err := GetFreePort()
	if err != nil {
		log.Fatal("fail to get freePort ")
	}
	startServe(freePort)
	node1.Name = "node1_org1"
	node1.Port = freePort
	node1.IP = "peer0.org1.example.com"
	node1.Host = "127.0.0.1"
	chainFlow := flow.MakeEmpFlow().Admin().Chain()

	// create chainFlow instance
	chainFlow.
		Identity().SetDecrypt(decrypt).Identity("ecdsa_with_sha256", privateKey, cert).
		TLS().SetRoots(roots2).SetDecrypt(decrypt).KeyPair(tlsKey, tlsCert).
		Net().AddNode(&node1)

	// chain operation
	err = chainFlow.SetTimeout(60).SetChain("1").SetGenesis(genesisPath).
		SetDecrypt(decryption).SetNode("node1_org1").CreateChain()
	assert.Equal(t, err, nil, "CreateChain Fail")
	_, err = chainFlow.SetTimeout(60).SetNode("node1_org1").QueryAllChain()
	assert.Equal(t, err, nil, "QueryAllChain Fail")
	_, err = chainFlow.SetTimeout(60).SetNode(node1.Name).QueryChain("1")
	assert.Equal(t, err, nil, "QueryChain Fail")
	err = chainFlow.SetTimeout(60).SetNode(node1.Name).SetChain("1").DeleteChain()
	assert.Equal(t, err, nil, "DeleteChain fail")
}

// mock server for testing tx operation
func TestSendTxOperation(t *testing.T) {
	// open mock server
	freePort, err := GetFreePort()
	if err != nil {
		log.Fatal("fail to get freePort ")
	}
	startServe(freePort)
	node1.Name = "node1_org1"
	node1.Port = freePort
	node1.IP = "peer0.org1.example.com"
	node1.Host = "127.0.0.1"
	flow := flow.MakeEmpFlow().Operator()
	os.Create("config/contract.zip")

	// create flow instance
	flow.
		Identity().SetDecrypt(decrypt).Identity("ecdsa_with_sha256", privateKey, cert).
		TLS().SetRoots(roots2).SetDecrypt(decrypt).KeyPair(tlsKey, tlsCert).
		Net().AddNode(&node1)

	// import-vote-send-query
	err = flow.Lifecycle().SetTimeout(60).SetContract("1", "test", "1").
		SetPkgPath(pkgPath).SetContractType("DockerGo").
		SetNode("node1_org1").Install("node1_org1")
	assert.Equal(t, err, nil, "Install fail")
	err = flow.Lifecycle().SetTimeout(60).SetConsenterList(consenter).
		SetContract("1", "test", "1").SetHistory(false).
		SetInit(false).Vote(description, policy)
	assert.Equal(t, err, nil, "Vote fail")
	userFlow := flow.User()
	err = userFlow.Contract().SetContractName(contractName).SetDomains(domain).SetChain("1").
		SetConsenterList(consenter).Invoke(method, args)
	if err != nil {
		log.Info(err.Error())
	}
	err = userFlow.Contract().SetContractName(contractName).SetDomains(domain).SetChain("1").
		SetNode("node1_org1").Query(method, args)
	if err != nil {
		log.Info(err.Error())
	}
}

func TestObserverFlow(t *testing.T) {
	node1.Name = "node1_org1"
	node1.Port = 12020
	node1.IP = "1.1.1.1"
	node1.Host = "1.1.1.1"

	observerFlow := flow.MakeEmpFlow().
		Identity().SetDecrypt(decrypt).Identity("ecdsa_with_sha256", privateKey, cert).
		TLS().SetRoots(roots1).SetDecrypt(decrypt).KeyPair(tlsKeyOther, tlsCertOther).
		Net().AddNode(&node1).Observer()

	observerFlow.SetNodeWithErr("node1_org1")
	observerFlow.SetNode("node1_org1").SetChain("1").QueryBlock(0) // 查询块高
	observerFlow.SetChain("1").SetNode("node1_org1").QueryTx(txHash)
	observerFlow.SetChain("1").SetNode("node1_org1").QueryTxResult(txHash)
	observerFlow.SetNode("node1_org1").SetChain("1").Height()
}

func TestLifeCycleFlow(t *testing.T) {
	node1.Name = "node1_org1"
	node1.Port = 12020
	node1.IP = "1.1.1.1"
	node1.Host = "1.1.1.1"
	os.Create("config/contract.zip")

	operatorFlow := flow.MakeEmpFlow().
		Identity().SetDecrypt(decrypt).Identity("ecdsa_with_sha256", privateKey, cert).
		TLS().SetRoots(roots1).SetDecrypt(decrypt).KeyPair(tlsKeyOther, tlsCertOther).
		Net().AddNode(&node1).Operator()

	operatorFlow.Lifecycle().SetContract("1", "test", "1").
		SetPkgPath(pkgPath).SetContractType("DockerGo").SetNode("node1_org1").
		Install("node1_org1")
	operatorFlow.Lifecycle().SetContract("1", "test", "1").
		SetContractType("DockerGo").SetNode("node1_org1").Uninstall()
	operatorFlow.Lifecycle().SetContract("1", "test", "1").
		SetChain("1").SetNode("node1_org1").QueryContractState("contract")
	operatorFlow.Lifecycle().SetContract("1", "test", "1").
		SetChain("1").SetNode("node1_org1").SetConsenterList(consenter).Freeze("contract")
	operatorFlow.Lifecycle().SetContract("1", "test", "1").
		SetHistory(false).SetInit(false).Vote(description, policy)
}

func TestContract(t *testing.T) {
	node1.Name = "node1_org1"
	node1.Port = 12020
	node1.IP = "1.1.1.1"
	node1.Host = "1.1.1.1"

	node2.Name = "node2_org1"
	node2.Port = 12020
	node2.IP = "1.1.1.1"
	node2.Host = "1.1.1.1"

	node3.Name = "node1_org2"
	node3.Port = 12020
	node3.IP = "1.1.1.1"
	node3.Host = "1.1.1.1"

	node4.Name = "node2_org2"
	node4.Port = 12020
	node4.IP = "1.1.1.1"
	node4.Host = "1.1.1.1"

	userFlow := flow.MakeEmpFlow().
		Identity().SetDecrypt(decrypt).Identity("ecdsa_with_sha256", privateKey, cert).
		TLS().SetRoots(roots1).SetDecrypt(decrypt).KeyPair(tlsKeyOther, tlsCertOther).
		Net().AddNode(&node1).
		User()

	userFlow.Contract().SetContractName(contractName).SetDomains(domain).
		SetChain("1").SetConsenterList(consenter).Invoke(method, args)
	userFlow.Contract().SetContractName(contractName).SetDomains(domain).
		SetChain("1").SetNode("node1_org1").Query(method, args)
}

func TestConfigFlow(t *testing.T) {
	node1.Name = "node1_org1"
	node1.Port = 12020
	node1.IP = "1.1.1.1"
	node1.Host = "1.1.1.1"

	operatorFlow := flow.MakeEmpFlow().
		Identity().SetDecrypt(decrypt).Identity("ecdsa_with_sha256", privateKey, cert).
		TLS().SetRoots(roots1).SetDecrypt(decrypt).KeyPair(tlsKeyOther, tlsCertOther).
		Net().AddNode(&node1).Operator()

	operatorFlow.Config().UpdatePolicy("1", policy)
	operatorFlow.Config().UpdateLifecycle("1", policy)
	operatorFlow.Config().SetNode("node1_org1").QueryChainVoteUpdate("1", subject)
}
