/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2022-2022. All rights reserved.
 */

package rawmessage

import (
	"regexp"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/util"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
	"huawei.com/huaweichain/proto/nodeservice"
	"huawei.com/huaweichain/sdk/crypto"
)

// CircleRawMessage is the definition of circle raw message.
type CircleRawMessage struct {
	builder MsgBuilder
	crypto  crypto.Crypto
}

const (
	// CircleNameReg is the circle name regex format.
	CircleNameReg = "^[A-Za-z0-9][A-Za-z0-9_-]{0,99}$"
)

// UpdateCircleRawMessage is used to create a raw message by specifying a message builder.
func UpdateCircleRawMessage(builder MsgBuilder, crypto crypto.Crypto) *CircleRawMessage {
	return &CircleRawMessage{builder: builder, crypto: crypto}
}

// BuildVoteRawMessage is used to build vote raw message for circle vote.
func (msg *CircleRawMessage) BuildVoteRawMessage(circleConfig *nodeservice.CircleConfig) (*TxRawMsg, error) {
	if err := checkCircleVoteParams(circleConfig); err != nil {
		return nil, errors.WithMessage(err, "check circle vote params error")
	}
	rawMsg, err := msg.buildVoteRawMessage(circleConfig)
	if err != nil {
		return nil, errors.WithMessage(err, "build vote raw message error")
	}
	return rawMsg, nil
}

func (msg *CircleRawMessage) buildVoteRawMessage(circleConfig *nodeservice.CircleConfig) (*TxRawMsg, error) {
	circleConfigDefinitionBytes, err := proto.Marshal(circleConfig)
	if err != nil {
		return nil, errors.WithMessage(err, "marshal circle definition error")
	}
	voteTxData := &common.VoteTxData{
		Handler: "circle_config",
		Payload: circleConfigDefinitionBytes,
	}
	tx, err := msg.builder.BuildVoteTx(circleConfig.ChainId, voteTxData)
	if err != nil {
		return nil, errors.WithMessage(err, "build vote tx error")
	}
	return msg.builder.GetTxRawMsg(tx)
}

// checkCircleVoteParams check the format of circle vote parameters
func checkCircleVoteParams(circleConfig *nodeservice.CircleConfig) error {
	if err := util.CheckChainID(circleConfig.ChainId); err != nil {
		return errors.WithMessage(err, "check chain id error")
	}
	if err := checkCircleName(circleConfig.Name); err != nil {
		return errors.WithMessage(err, "check circle name error")
	}
	if err := util.CheckPolicy(circleConfig.EndorsementPolicy); err != nil {
		return errors.WithMessage(err, "check policy error")
	}
	if err := checkPeer(circleConfig.RequiredPeer, circleConfig.MaximumPeer); err != nil {
		return errors.WithMessage(err, "check peer error")
	}
	if err := checkCircleDBType(circleConfig.DbType); err != nil {
		return errors.WithMessage(err, "check db type error")
	}
	if err := checkCircleOrgs(circleConfig.Organizations); err != nil {
		return errors.WithMessage(err, "check organizations error")
	}
	return nil
}

// checkCircleName check the circle name format
func checkCircleName(name string) error {
	r, err := regexp.Compile(CircleNameReg)
	if err != nil {
		return errors.WithMessage(err, "regexp compile error")
	}
	if !r.MatchString(name) {
		return errors.Errorf("check circle name not match format: %s", CircleNameReg)
	}
	return nil
}

// checkCircleName check the circle name format
func checkPeer(requiredPeer int32, maxPeer int32) error {
	if maxPeer < requiredPeer {
		return errors.New("circleInfo maximum peer count less than required")
	}
	return nil
}

func checkCircleDBType(dbType nodeservice.DBType) error {
	if nodeservice.DBType_name[int32(dbType)] == "" {
		return errors.Errorf("the circle dbType is %s", dbType)
	}
	return nil
}

func checkCircleOrgs(orgs []*common.Organization) error {
	if len(orgs) == 0 {
		return errors.New("empty organizations")
	}
	for _, org := range orgs {
		if org.Name == "" {
			return errors.New("the circle org name is empty")
		}
	}
	return nil
}
