/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

// Package utils provide the common utils.
package utils

import (
	"crypto/rand"
	"crypto/sha256"
	"fmt"
	"math"
	"math/big"
	mrand "math/rand"
	"strconv"
	"time"

	"github.com/pkg/errors"

	"huawei.com/huaweichain/common/cryptomgr/gmalg"
	"huawei.com/huaweichain/proto"
	"huawei.com/huaweichain/proto/common"
)

const (
	// Seed is the random factor for generating tx id.
	Seed = 10000
	// ShardChainIDSep is used to separate chain id and shard id.
	ShardChainIDSep = "-"
)

var timeout time.Duration = 60

// HashSha256 computes hash by sha256 algorithm.
func HashSha256(data []byte) []byte {
	h := sha256.Sum256(data)
	return h[:]
}

// HashSM3 computes hash by sm3 algorithm.
func HashSM3(data []byte) []byte {
	return gmalg.Hash(data)
}

// GenerateTimestamp is used to generate timestamp.
func GenerateTimestamp() uint64 {
	return uint64(time.Now().UTC().UnixNano())
}

// GenerateNonce generates a random number.
func GenerateNonce() uint64 {
	n, err := rand.Int(rand.Reader, big.NewInt(math.MaxInt64))
	if err != nil {
		mrand.Seed(time.Now().UnixNano())
		return mrand.Uint64() // nolint:gosec // this is just a backup operation for secure random
	}
	return n.Uint64()
}

// GetShards used to parse shard id from invoke response.
func GetShards(tx *common.Transaction) ([]uint32, error) {
	var txPayLoad common.TxPayload
	if err := proto.Unmarshal(tx.Payload, &txPayLoad); err != nil {
		return nil, errors.WithMessage(err, "tx payload unmarshal error")
	}
	var txData common.CommonTxData
	if err := proto.Unmarshal(txPayLoad.Data, &txData); err != nil {
		return nil, errors.WithMessage(err, "common tx data unmarshal error")
	}
	var shardIds []uint32
	for _, shardUpdate := range txData.ShardStateUpdates {
		shardIds = append(shardIds, shardUpdate.ShardId)
	}
	return shardIds, nil
}

// GetShardChainID is used to construct shard chain id with origin chain id and shard id.
func GetShardChainID(chainID string, shardID uint32) string {
	return fmt.Sprintf("%s%s%s", chainID, ShardChainIDSep, strconv.Itoa(int(shardID)))
}

// SetTimeout set client timeout (/s)
func SetTimeout(seconds time.Duration) {
	timeout = seconds
}

// GetTimeout get client timeout (/s)
func GetTimeout() time.Duration {
	return timeout
}
